#include <stdio.h>
#include <d3dx9.h>
#include "file_ect.h"
#include "light.h"

#pragma comment(lib,"d3d9.lib")
#pragma comment(lib,"d3dx9.lib")

#define MYWNDTITLE		"DX9_FileEct"
#define MYWNDCLASS		"CLASS_DX9_FileEct"
#define WIDTH			640 			// ʂ̉TCY
#define HEIGHT			480 			// ʂ̏cTCY

LPDIRECT3D9 			g_pD3D = NULL;
LPDIRECT3DDEVICE9		g_pD3DDev = NULL;
D3DPRESENT_PARAMETERS	g_d3dpp;
CFileECT	g_meshEct;
ANIMSET 	g_as;
int 		g_idx_anim = 0;
int 		g_idx_motion = 0;
float		g_roll = 0;

HRESULT InitD3D(HWND hWnd);
void	InitGeometry();
void	SetupMatrices();
void	Set_Light();
void	Draw_Main();
void	Get_Key();
void	Motion_Count();
void	Set_FullPath(char *path, char *filename);
void	Cleanup();

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_CREATE:
		{
			if(FAILED(InitD3D(hWnd))){
				MessageBox(hWnd,"D3D9̏Ɏs܂",MYWNDTITLE,0);
				SendMessage(hWnd,WM_CLOSE,0,0);
				return 0;
			}
			// ^C}[ݒ
			SetTimer(hWnd, 1, 100, NULL);
		}
		break;
		
	case WM_TIMER:
		Get_Key();		// L[͎擾
		Draw_Main();	// IuWFNg`
		break;
		
	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_SPACE:	// [Space] L[ƃAj؂ւ
			if(&g_meshEct){
				int anim_ct = g_meshEct.GetAnimCount();
				if(anim_ct > 0){
					g_idx_anim ++;
					if(g_idx_anim >= anim_ct)
						g_idx_anim = 0;
					g_as = g_meshEct.GetAnimSet(g_idx_anim);
					g_idx_motion = g_as.iStart;
					g_meshEct.SetMotionIndex(g_idx_motion);
				}
			}
			break;
		case VK_ESCAPE: // [ESC] L[ƏI
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			break;
		}
		break;
		
	case WM_CLOSE:
		KillTimer(hWnd,1);		// ^C}[
		DestroyWindow(hWnd);
		Cleanup();				// Direct3D IuWFNg
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
	return 0;
}

int WINAPI WinMain(HINSTANCE hInst,HINSTANCE hPrevInst,LPSTR lpCmdLine,int nCmdShow)
{
	WNDCLASSEX	wc;
	HWND		hWnd;

	ZeroMemory(&wc, sizeof(wc));
	wc.cbSize			= sizeof(WNDCLASSEX); 
	wc.lpfnWndProc		= (WNDPROC)WndProc;
	wc.hInstance		= hInst;
	wc.hIcon			= LoadIcon(hInst,IDI_WINLOGO);
	wc.hCursor			= LoadCursor(NULL, MAKEINTRESOURCE(IDC_ARROW));
	wc.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	wc.lpszClassName	= MYWNDCLASS;
	if(RegisterClassEx(&wc) == 0)
		return 0;
	
	DWORD	dwStyle = WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX | WS_CAPTION | WS_VISIBLE;
	RECT	rc = {0, 0, WIDTH, HEIGHT};
	AdjustWindowRectEx(&rc, dwStyle, FALSE, NULL);
	
	hWnd = CreateWindowEx(0, 
				MYWNDCLASS,
				MYWNDTITLE,
				dwStyle,
				CW_USEDEFAULT, CW_USEDEFAULT, rc.right - rc.left, rc.bottom - rc.top,
				NULL,
				NULL,
				hInst,
				NULL);
	
	if(hWnd == NULL)
		return 0;

	MSG msg; 
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}

// D3D9
HRESULT InitD3D(HWND hWnd)
{
	// Direct3D9 쐬
	g_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if (!g_pD3D){
		return E_FAIL;
	}
	
	ZeroMemory(&g_d3dpp, sizeof(g_d3dpp));
	g_d3dpp.Windowed = TRUE;
	g_d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	g_d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	g_d3dpp.EnableAutoDepthStencil = TRUE;
	g_d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	g_d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	
	// D3D9Device 쐬
	if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
									  D3DCREATE_HARDWARE_VERTEXPROCESSING,
									  &g_d3dpp, &g_pD3DDev))){
		if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
										  D3DCREATE_SOFTWARE_VERTEXPROCESSING,
										  &g_d3dpp, &g_pD3DDev))){
			if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, hWnd,
										  D3DCREATE_SOFTWARE_VERTEXPROCESSING,
										  &g_d3dpp, &g_pD3DDev))){
				return E_FAIL;
			}
		}
	}
	
	// ݒ
	InitGeometry();
	
	return S_OK;
}//InitD3D

// ݒ菉
void InitGeometry()
{
	// JOݒ
	g_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
	// Zobt@L
	g_pD3DDev->SetRenderState(D3DRS_ZENABLE, TRUE);
	// eNX` tB^ݒ
	g_pD3DDev->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	g_pD3DDev->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	char path[MAX_PATH];
	// ECTt@Cǂݍ
	Set_FullPath(path, "robot2.ect");
	g_meshEct.Load(g_pD3DDev, path);
	// ECTt@C̃[Vԓǂݍ
	g_as = g_meshEct.GetAnimSet(g_idx_anim);
	// Cgݒ
	Set_Light();
	// ϊsݒ
	SetupMatrices();
}// InitGeometry

// ϊsݒ
void SetupMatrices()
{
	// r[s
	D3DXMATRIX	matView;
	D3DXVECTOR3 look(0,0,-2.5f);
	D3DXVECTOR3 lookat(0,0,0);
	D3DXVECTOR3 lookup(0,1,0);
	D3DXMatrixLookAtLH(&matView, &look, &lookat, &lookup);
	g_pD3DDev->SetTransform(D3DTS_VIEW,&matView);
	// vWFNVs
	D3DXMATRIX	matProj;
	float		view_rad = D3DX_PI / 4;
	float		aspect	 = (float) WIDTH / HEIGHT;
	float		nearest  = 0.01f;
	float		farthest = 10.0f;
	D3DXMatrixPerspectiveFovLH(&matProj, view_rad, aspect, nearest, farthest);
	g_pD3DDev->SetTransform(D3DTS_PROJECTION,&matProj);
}// SetupMatrices

// Ɩݒ
void Set_Light()
{
	CLight				light;
	LIGHT_DIRECTIONAL	inf;
	
	inf.Color.r = inf.Color.g = inf.Color.b = 1.0f;
	D3DXVec3Normalize((D3DXVECTOR3*) &inf.Direction, &D3DXVECTOR3(1,-2,1));
	light.SetFlagLight(g_pD3DDev, 0, TRUE);
	light.SetLightDirectional(g_pD3DDev, 0, inf);
	light.SetAmbient(g_pD3DDev, D3DCOLOR_XRGB(100,100,100));
	light.SetFlagSpecular(g_pD3DDev, TRUE);
}// Set_Light

// `
void Draw_Main()
{
	// obNobt@ɕ`
	g_pD3DDev->Clear(0,NULL,D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,D3DCOLOR_XRGB(0,0,0),1.0f,0);
	if(SUCCEEDED(g_pD3DDev->BeginScene()))
	{
		D3DXMATRIX	matRY;
		D3DXMatrixRotationY(&matRY, g_roll/180*D3DX_PI);
		g_pD3DDev->SetTransform(D3DTS_WORLD, &matRY);
		
		g_meshEct.Render(g_pD3DDev);
		
		g_pD3DDev->EndScene();
		g_pD3DDev->SetTexture(0, NULL);
	}
	g_pD3DDev->Present(NULL,NULL,NULL,NULL);
	
	// foCXƂ
	if(g_pD3DDev->TestCooperativeLevel() == D3DERR_DEVICENOTRESET){
		g_pD3DDev->Reset(&g_d3dpp);
		InitGeometry();
		Draw_Main();
	}
}// Draw_Main

// L[
void Get_Key()
{
	if(GetAsyncKeyState(VK_LEFT) & 0x8000){ 		// []L[ꂽƂ
		g_roll -= 5;
		if(g_roll <= 0)
			g_roll += 360;
	}
	else if(GetAsyncKeyState(VK_RIGHT) & 0x8000){	// []L[ꂽƂ
		g_roll += 5;
		if(g_roll >= 360)
			g_roll -= 360;
	}
	
	if(GetAsyncKeyState(VK_UP) & 0x8000){			// []L[ꂽƂ
		Motion_Count();
	}
	else if(GetAsyncKeyState(VK_DOWN) & 0x8000){	// []L[ꂽƂ
		Motion_Count();
	}
}// Get_Key

// \郂[V̔ԍvZ
void Motion_Count()
{
	g_idx_motion ++;
	
	if(g_idx_motion > g_as.iEnd){
		g_idx_motion = g_as.iStart;
	}
	g_meshEct.SetMotionIndex(g_idx_motion);
}// Motion_Count

// t@CɎst@C̃pXt
// path 	: [OUT] tpX
// filename : [IN]	t@C
void Set_FullPath(char *path, char *filename)
{
	GetModuleFileName(NULL, path, MAX_PATH);
	char *p = strrchr(path, '\\');
	if(p+1)
		strcpy(p+1, filename);
}// Set_FullPath

// IuWFNg
void Cleanup()
{
	// Direct3D IuWFNg
	SAFE_RELEASE(g_pD3DDev);
	SAFE_RELEASE(g_pD3D);
}// Cleanup
