// `bV
#ifndef _BOARD_H_
#define _BOARD_H_

#include <d3dx9.h>
#include "macro.h"
#include "fvf.h"
#include "fvf_rhw.h"

class CBoard
{
public:
	CBoard();
	~CBoard();
	void	Free();
	void	Init(DWORD dwFVF, float x_min, float x_max, float y_min, float y_max,
				float z_min, float z_max);
	void	Resize(float x_min, float x_max, float y_min, float y_max,
					float z_min, float z_max);
	HRESULT LoadTexture(LPDIRECT3DDEVICE9 pD3DDev, char *strTextureFilename,
						D3DCOLOR colorkey = 0);
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev, D3DCOLOR color = 0xffffffff);
	D3DMATERIAL9			GetMaterial() { return *m_pMaterial; }
	void					SetMaterial(D3DMATERIAL9 *pMtrl) { m_pMaterial = pMtrl; }
	LPDIRECT3DTEXTURE9		GetTexture() { return m_pTex; }
	
private:
	DWORD				m_dwFVF;
	D3DXVECTOR3 		m_pos[4];
	D3DXVECTOR2 		m_texcoord[4];
	D3DXVECTOR3 		m_normal;
	D3DMATERIAL9*		m_pMaterial;	// }eA
	LPDIRECT3DTEXTURE9	m_pTex; 		// eNX`
	
private:
	D3DXVECTOR3 GetFaceNormal(D3DXVECTOR3 pos[4]);
	void SetVertexNT(VERTEX_NT *pv, D3DXVECTOR3 pos, D3DXVECTOR3 n, D3DXVECTOR2 tex);
	void SetVertexCT(VERTEX_CT *pv, D3DXVECTOR3 pos, D3DCOLOR color,
					D3DXVECTOR2 tex);
	void SetVertexT(VERTEX_T *pv, D3DXVECTOR3 pos, D3DXVECTOR2 tex);
	void SetVertexN(VERTEX_N *pv, D3DXVECTOR3 pos, D3DXVECTOR3 n);
	void SetVertexC(VERTEX_C *pv, D3DXVECTOR3 pos, D3DCOLOR color);
	void SetVertexRHWT(VERTEX_RHWT *pv, D3DXVECTOR3 pos, D3DXVECTOR2 tex);
	void SetVertexRHWCT(VERTEX_RHWCT *pv, D3DXVECTOR3 pos, D3DCOLOR color,
						D3DXVECTOR2 tex);
	void SetVertexRHWC(VERTEX_RHWC *pv, D3DXVECTOR3 pos, D3DCOLOR color);
};

CBoard::CBoard()
{
	m_dwFVF = 0;
	ZeroMemory(m_pos, sizeof(m_pos));
	ZeroMemory(m_texcoord, sizeof(m_texcoord));
	ZeroMemory(m_normal, sizeof(m_normal));
	m_pMaterial = NULL;
	m_pTex = NULL;
}

CBoard::~CBoard()
{
	Free();
}

void CBoard::Free()
{
	SAFE_RELEASE(m_pTex);
}

// r{[h쐬
// dwFVF	   : [IN] _tH[}bg
// x_min,x_max : [IN] XW̍ŏlƍől
// y_min,y_max : [IN] YW̍ŏlƍől
// z_min,z_max : [IN] ZW̍ŏlƍől
void CBoard::Init(DWORD dwFVF, float x_min, float x_max, float y_min, float y_max,
					float z_min, float z_max)
{
	m_dwFVF = dwFVF;
	Resize(x_min, x_max, y_min, y_max, z_min, z_max);
	m_texcoord[0] = D3DXVECTOR2(0.0f, 0.0f);
	m_texcoord[1] = D3DXVECTOR2(1.0f, 0.0f);
	m_texcoord[2] = D3DXVECTOR2(1.0f, 1.0f);
	m_texcoord[3] = D3DXVECTOR2(0.0f, 1.0f);
	if(m_dwFVF & D3DFVF_NORMAL){
		m_normal = GetFaceNormal(m_pos);
	}
}

// TCYύX
// x_min,x_max : [IN] XW̍ŏlƍől
// y_min,y_max : [IN] YW̍ŏlƍől
// z_min,z_max : [IN] ZW̍ŏlƍől
void CBoard::Resize(float x_min, float x_max, float y_min, float y_max,
					float z_min, float z_max)
{
	if(!(m_dwFVF & D3DFVF_XYZRHW)){
		m_pos[0] = D3DXVECTOR3(x_min, y_max, z_max);
		m_pos[1] = D3DXVECTOR3(x_max, y_max, z_max);
		m_pos[2] = D3DXVECTOR3(x_max, y_min, z_min);
		m_pos[3] = D3DXVECTOR3(x_min, y_min, z_min);
	}
	else{
		m_pos[0] = D3DXVECTOR3(x_min, y_min, z_min);
		m_pos[1] = D3DXVECTOR3(x_max, y_min, z_min);
		m_pos[2] = D3DXVECTOR3(x_max, y_max, z_max);
		m_pos[3] = D3DXVECTOR3(x_min, y_max, z_max);
	}
}

HRESULT CBoard::LoadTexture(LPDIRECT3DDEVICE9 pD3DDev, char *strTextureFilename,
							D3DCOLOR colorkey)
{
	if(!pD3DDev)
		return E_FAIL;
	
	SAFE_RELEASE(m_pTex);
	
	HRESULT hr = E_FAIL;
	
	if(strTextureFilename){
		hr = D3DXCreateTextureFromFileEx(pD3DDev, strTextureFilename, 0, 0, 0, 0,
								D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, D3DX_FILTER_LINEAR,
								D3DX_FILTER_LINEAR, colorkey, NULL, NULL, &m_pTex);
	}
	
	return hr;
}

// _O
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
// color   : [IN] _̃fBt[YF
HRESULT CBoard::Render(LPDIRECT3DDEVICE9 pD3DDev, D3DCOLOR color)
{
	if(!pD3DDev)
		return E_FAIL;
	
	int 					i;
	HRESULT 				hr;
	LPDIRECT3DVERTEXSHADER9 dwVS_pre;
	DWORD					dwFVF_pre,dwLighting_pre;
	
	hr = pD3DDev->GetVertexShader(&dwVS_pre);
	hr = pD3DDev->GetFVF(&dwFVF_pre);
	hr = pD3DDev->GetRenderState(D3DRS_LIGHTING, &dwLighting_pre);
	
	hr = pD3DDev->SetVertexShader(NULL);
	if(m_dwFVF & D3DFVF_NORMAL){
		hr = pD3DDev->SetRenderState(D3DRS_LIGHTING, TRUE);
	}
	else{
		hr = pD3DDev->SetRenderState(D3DRS_LIGHTING, FALSE);
	}
	
	if(m_pTex)
		hr = pD3DDev->SetTexture(0, m_pTex);
	if(m_pMaterial)
		hr = pD3DDev->SetMaterial(m_pMaterial);
	
	switch (m_dwFVF)
	{
	case FVF_VERTEX_NT:
		{
			VERTEX_NT v[4];
			for(i=0; i<4; i++){
				SetVertexNT(&v[i], m_pos[i], m_normal, m_texcoord[i]);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_NT);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_NT));
		}
		break;
	case FVF_VERTEX_CT:
		{
			VERTEX_CT v[4];
			for(i=0; i<4; i++){
				SetVertexCT(&v[i], m_pos[i], color, m_texcoord[i]);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_CT);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_CT));
		}
		break;
	case FVF_VERTEX_T:
		{
			VERTEX_T v[4];
			for(i=0; i<4; i++){
				SetVertexT(&v[i], m_pos[i], m_texcoord[i]);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_T);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_T));
		}
		break;
	case FVF_VERTEX_N:
		{
			VERTEX_N v[4];
			for(i=0; i<4; i++){
				SetVertexN(&v[i], m_pos[i], m_normal);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_N);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_N));
		}
		break;
	case FVF_VERTEX_C:
		{
			VERTEX_C v[4];
			for(i=0; i<4; i++){
				SetVertexC(&v[i], m_pos[i], color);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_C);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_C));
		}
		break;
	case FVF_VERTEX_RHWCT:
		{
			VERTEX_RHWCT v[4];
			for(i=0; i<4; i++){
				SetVertexRHWCT(&v[i], m_pos[i], color, m_texcoord[i]);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_RHWCT);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_RHWCT));
		}
		break;
	case FVF_VERTEX_RHWT:
		{
			VERTEX_RHWT v[4];
			for(i=0; i<4; i++){
				SetVertexRHWT(&v[i], m_pos[i], m_texcoord[i]);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_RHWT);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_RHWT));
		}
		break;
	case FVF_VERTEX_RHWC:
		{
			VERTEX_RHWC v[4];
			for(i=0; i<4; i++){
				SetVertexRHWC(&v[i], m_pos[i], color);
			}
			hr = pD3DDev->SetFVF(FVF_VERTEX_RHWC);
			hr = pD3DDev->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, v, sizeof(VERTEX_RHWC));
		}
		break;
	}
	
	hr = pD3DDev->SetVertexShader(dwVS_pre);
	hr = pD3DDev->SetFVF(dwFVF_pre);
	hr = pD3DDev->SetRenderState(D3DRS_LIGHTING, dwLighting_pre);
	
	return hr;
}

// ʖ@xNg擾
D3DXVECTOR3 CBoard::GetFaceNormal(D3DXVECTOR3 pos[4])
{
	D3DXVECTOR3 normal,normal2,normal3;
	D3DXVec3Cross(&normal2, &(pos[1] - pos[0]), &(pos[2] - pos[1]));
	D3DXVec3Cross(&normal3, &(pos[2] - pos[1]), &(pos[3] - pos[2]));
	D3DXVec3Normalize(&normal, &(normal2 + normal3));
	
	return normal;
}

// VERTEX_NT \̂ɒlݒ
void CBoard::SetVertexNT(VERTEX_NT *pv, D3DXVECTOR3 pos, D3DXVECTOR3 n,
						D3DXVECTOR2 tex)
{
	pv->v	= pos;
	pv->n	= n;
	pv->tu	= tex.x;
	pv->tv	= tex.y;
}

// VERTEX_CT \̂ɒlݒ
void CBoard::SetVertexCT(VERTEX_CT *pv, D3DXVECTOR3 pos, D3DCOLOR color,
						D3DXVECTOR2 tex)
{
	pv->v	= pos;
	pv->c	= color;
	pv->tu	= tex.x;
	pv->tv	= tex.y;
}

// VERTEX_T \̂ɒlݒ
void CBoard::SetVertexT(VERTEX_T *pv, D3DXVECTOR3 pos, D3DXVECTOR2 tex)
{
	pv->v	= pos;
	pv->tu	= tex.x;
	pv->tv	= tex.y;
}

// VERTEX_N \̂ɒlݒ
void CBoard::SetVertexN(VERTEX_N *pv, D3DXVECTOR3 pos, D3DXVECTOR3 n)
{
	pv->v	= pos;
	pv->n	= n;
}

// VERTEX_C \̂ɒlݒ
void CBoard::SetVertexC(VERTEX_C *pv, D3DXVECTOR3 pos, D3DCOLOR color)
{
	pv->v	= pos;
	pv->c	= color;
}

// VERTEX_RHWT \̂ɒlݒ
void CBoard::SetVertexRHWT(VERTEX_RHWT *pv, D3DXVECTOR3 pos, D3DXVECTOR2 tex)
{
	pv->v	= pos;
	pv->rhw = 1.0f;
	pv->tu	= tex.x;
	pv->tv	= tex.y;
}

// VERTEX_RHWCT \̂ɒlݒ
void CBoard::SetVertexRHWCT(VERTEX_RHWCT *pv, D3DXVECTOR3 pos, D3DCOLOR color,
						D3DXVECTOR2 tex)
{
	pv->v	= pos;
	pv->rhw = 1.0f;
	pv->c	= color;
	pv->tu	= tex.x;
	pv->tv	= tex.y;
}

// VERTEX_RHWC \̂ɒlݒ
void CBoard::SetVertexRHWC(VERTEX_RHWC *pv, D3DXVECTOR3 pos, D3DCOLOR color)
{
	pv->v	= pos;
	pv->rhw = 1.0f;
	pv->c	= color;
}

#endif