// ov
// l^FDirectX9 Tv
#ifndef _BUMP_ENV_H_
#define _BUMP_ENV_H_

#include <d3dx9.h>
#include "macro.h"
#include "fvf.h"

class CBumpEnv
{
public:
	CBumpEnv();
	virtual ~CBumpEnv();
	void	Free();
	HRESULT Init(LPDIRECT3DDEVICE9 pD3DDev,
				char *strBaseTextureFilename,
				char *strEnvMapTextureFilename,
				char *strHeightMapTextureFilename);
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev);
	virtual HRESULT RenderMesh(LPDIRECT3DDEVICE9 pD3DDev) = 0;
	HRESULT InitBumpMap(LPDIRECT3DDEVICE9 pD3DDev);
	
private:
	LPDIRECT3DTEXTURE9	m_pBaseTexture; 		// {eNX`
	LPDIRECT3DTEXTURE9	m_pEnvMapTexture;		// }bveNX`
	LPDIRECT3DTEXTURE9	m_pHeightMapTexture;	// wCg}bveNX`
	LPDIRECT3DTEXTURE9	m_psBumpMap;			// ov}bveNX`
	
private:
	HRESULT CreateWaterVB( LPDIRECT3DDEVICE9 pD3DDev );
	LPDIRECT3DTEXTURE9 CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev,
			DWORD dwWidth, DWORD dwHeight, D3DFORMAT d3dBumpFormat );
};

CBumpEnv::CBumpEnv()
{
	m_pBaseTexture		= NULL;
	m_pEnvMapTexture	= NULL;
	m_pHeightMapTexture = NULL;
	m_psBumpMap 		= NULL;
}

CBumpEnv::~CBumpEnv()
{
	Free();
}

void CBumpEnv::Free()
{
	SAFE_RELEASE(m_pBaseTexture);
	SAFE_RELEASE(m_pEnvMapTexture);
	SAFE_RELEASE(m_pHeightMapTexture);
	SAFE_RELEASE(m_psBumpMap);
}

// 
// pD3DDev					   : [IN] IDirect3DDevice9 IuWFNg
// strBaseTextureFilename	   : [IN] {eNX` t@C
// strEnvMapTextureFilename    : [IN] }bveNX` t@C
// strHeightMapTextureFilename : [IN] wCg}bveNX` t@C
HRESULT CBumpEnv::Init(LPDIRECT3DDEVICE9 pD3DDev,
						char *strBaseTextureFilename,
						char *strEnvMapTextureFilename,
						char *strHeightMapTextureFilename)
{
	if(!pD3DDev)
		return E_FAIL;
	
	Free();
	
	HRESULT hr;
	
	// eNX`쐬
	hr = CreateTexture( pD3DDev, strBaseTextureFilename,
						&m_pBaseTexture, D3DFMT_R5G6B5 );
	hr = CreateTexture( pD3DDev, strEnvMapTextureFilename,
						&m_pEnvMapTexture, D3DFMT_R5G6B5 );
	hr = CreateTexture( pD3DDev, strHeightMapTextureFilename,
						&m_pHeightMapTexture, D3DFMT_A8R8G8B8 );
	
	// ov}bv
	InitBumpMap(pD3DDev);
	
	return hr;
}

// _O
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpEnv::Render(LPDIRECT3DDEVICE9 pD3DDev)
{
	if(!pD3DDev || !m_pBaseTexture || !m_psBumpMap || !m_pEnvMapTexture)
		return E_FAIL;
	
	HRESULT hr;
	DWORD	dwTex0Mag_pre,dwTex0Min_pre,dwTex0U_pre,dwTex0V_pre;
	DWORD	dwTex1Mag_pre,dwTex1Min_pre,dwTex1U_pre,dwTex1V_pre;
	DWORD	dwTex2Mag_pre,dwTex2Min_pre,dwTex2U_pre,dwTex2V_pre;
	DWORD	dwClr0Op_pre,dwClr0Arg1_pre,dwClr0Arg2_pre;
	DWORD	dwClr1Op_pre,dwClr1Arg1_pre,dwClr1Arg2_pre;
	DWORD	dwClr2Op_pre,dwClr2Arg1_pre,dwClr2Arg2_pre;
	DWORD	dwAlp0Op_pre,dwAlp0Arg1_pre;
	DWORD	dwBump0Mat00_pre,dwBump0Mat01_pre,dwBump0Mat10_pre,dwBump0Mat11_pre;
	DWORD	dwBump0Scale_pre,dwBump0Offset_pre;
	DWORD	dwTex0Index_pre,dwTex1Index_pre,dwTex2Index_pre;
	LPDIRECT3DTEXTURE9	pTex0_pre,pTex1_pre,pTex2_pre;
	
	// ȑȌԂێ
	hr = pD3DDev->GetTexture( 0, (LPDIRECT3DBASETEXTURE9*) &pTex0_pre );
	hr = pD3DDev->GetTexture( 1, (LPDIRECT3DBASETEXTURE9*) &pTex1_pre );
	hr = pD3DDev->GetTexture( 2, (LPDIRECT3DBASETEXTURE9*) &pTex2_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MAGFILTER, &dwTex0Mag_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MINFILTER, &dwTex0Min_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_ADDRESSU,  &dwTex0U_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_ADDRESSV,  &dwTex0V_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MAGFILTER, &dwTex1Mag_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MINFILTER, &dwTex1Min_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_ADDRESSU,  &dwTex1U_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_ADDRESSV,  &dwTex1V_pre );
	hr = pD3DDev->GetSamplerState( 2, D3DSAMP_MAGFILTER, &dwTex2Mag_pre );
	hr = pD3DDev->GetSamplerState( 2, D3DSAMP_MINFILTER, &dwTex2Min_pre );
	hr = pD3DDev->GetSamplerState( 2, D3DSAMP_ADDRESSU,  &dwTex2U_pre );
	hr = pD3DDev->GetSamplerState( 2, D3DSAMP_ADDRESSV,  &dwTex2V_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, &dwTex0Index_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, &dwTex1Index_pre );
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_TEXCOORDINDEX, &dwTex2Index_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLOROP,	 &dwClr0Op_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG1, &dwClr0Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG2, &dwClr0Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLOROP,	 &dwClr1Op_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG1, &dwClr1Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG2, &dwClr1Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_COLOROP,	 &dwClr2Op_pre );
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_COLORARG1, &dwClr2Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_COLORARG2, &dwClr2Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_ALPHAOP,	 &dwAlp0Op_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_ALPHAARG1, &dwAlp0Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT00,   &dwBump0Mat00_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT01,   &dwBump0Mat01_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT10,   &dwBump0Mat10_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT11,   &dwBump0Mat11_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLSCALE,  &dwBump0Scale_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, &dwBump0Offset_pre );
	
	// eNX`ݒ
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSU,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSV,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSU,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSV,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_ADDRESSU,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_ADDRESSV,  D3DTADDRESS_WRAP );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetTexture( 0, m_pBaseTexture );
	hr = pD3DDev->SetTexture( 1, m_psBumpMap );
	hr = pD3DDev->SetTexture( 2, m_pEnvMapTexture );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0 );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP,	 D3DTOP_SELECTARG1);
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP,	 D3DTOP_SELECTARG1 );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, 0 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,	 D3DTOP_BUMPENVMAP );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_CURRENT );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVMAT00,   F2DW(0.5f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVMAT01,   F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVMAT10,   F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVMAT11,   F2DW(0.5f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVLSCALE,  F2DW(4.0f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_BUMPENVLOFFSET, F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_SPHEREMAP );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLOROP,	 D3DTOP_ADD );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLORARG2, D3DTA_CURRENT );
	
	hr = RenderMesh(pD3DDev);
	
	// ɖ߂
	hr = pD3DDev->SetTexture( 0, pTex0_pre );
	hr = pD3DDev->SetTexture( 1, pTex1_pre );
	hr = pD3DDev->SetTexture( 2, pTex2_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, dwTex0Mag_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, dwTex0Min_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSU,  dwTex0U_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSV,  dwTex0V_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, dwTex1Mag_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, dwTex1Min_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSU,  dwTex1U_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSV,  dwTex1V_pre );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_MAGFILTER, dwTex2Mag_pre );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_MINFILTER, dwTex2Min_pre );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_ADDRESSU,  dwTex2U_pre );
	hr = pD3DDev->SetSamplerState( 2, D3DSAMP_ADDRESSV,  dwTex2V_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, dwTex0Index_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, dwTex1Index_pre );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_TEXCOORDINDEX, dwTex2Index_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP,	 dwClr0Op_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, dwClr0Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, dwClr0Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,	 dwClr1Op_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, dwClr1Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, dwClr1Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLOROP,	 dwClr2Op_pre );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLORARG1, dwClr2Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLORARG2, dwClr2Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP,	 dwAlp0Op_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, dwAlp0Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00,   dwBump0Mat00_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01,   dwBump0Mat01_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10,   dwBump0Mat10_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11,   dwBump0Mat11_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLSCALE,  dwBump0Scale_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, dwBump0Offset_pre );
	
	return hr;
}

// ov}bv
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpEnv::InitBumpMap(LPDIRECT3DDEVICE9 pD3DDev)
{
	LPDIRECT3DTEXTURE9	psBumpSrc = m_pHeightMapTexture;
	D3DSURFACE_DESC 	d3dsd;
	D3DLOCKED_RECT		d3dlr;
	D3DFORMAT			BumpMapFormat = D3DFMT_V8U8;

	psBumpSrc->GetLevelDesc( 0, &d3dsd );
	// Create the bumpmap's surface and texture objects
	if( FAILED( pD3DDev->CreateTexture( d3dsd.Width, d3dsd.Height, 1, 0,
		BumpMapFormat, D3DPOOL_MANAGED, &m_psBumpMap, NULL ) ) )
	{
		return E_FAIL;
	}

	// Fill the bits of the new texture surface with bits from
	// a private format.
	psBumpSrc->LockRect( 0, &d3dlr, 0, 0 );
	DWORD dwSrcPitch = (DWORD)d3dlr.Pitch;
	BYTE* pSrcTopRow = (BYTE*)d3dlr.pBits;
	BYTE* pSrcCurRow = pSrcTopRow;
	BYTE* pSrcBotRow = pSrcTopRow + (dwSrcPitch * (d3dsd.Height - 1) );

	m_psBumpMap->LockRect( 0, &d3dlr, 0, 0 );
	DWORD dwDstPitch = (DWORD)d3dlr.Pitch;
	BYTE* pDstTopRow = (BYTE*)d3dlr.pBits;
	BYTE* pDstCurRow = pDstTopRow;

	for( DWORD y=0; y<d3dsd.Height; y++ )
	{
		BYTE* pSrcB0; // addr of current pixel
		BYTE* pSrcB1; // addr of pixel below current pixel, wrapping to top if necessary
		BYTE* pSrcB2; // addr of pixel above current pixel, wrapping to bottom if necessary
		BYTE* pDstT;  // addr of dest pixel;

		pSrcB0 = pSrcCurRow;

		if( y == d3dsd.Height - 1)
			pSrcB1 = pSrcTopRow;
		else
			pSrcB1 = pSrcCurRow + dwSrcPitch;

		if( y == 0 )
			pSrcB2 = pSrcBotRow;
		else
			pSrcB2 = pSrcCurRow - dwSrcPitch;

		pDstT = pDstCurRow;

		for( DWORD x=0; x<d3dsd.Width; x++ )
		{
			LONG v00; // Current pixel
			LONG v01; // Pixel to the right of current pixel, wrapping to left edge if necessary
			LONG vM1; // Pixel to the left of current pixel, wrapping to right edge if necessary
			LONG v10; // Pixel one line below.
			LONG v1M; // Pixel one line above.

			v00 = *(pSrcB0+0);

			if( x == d3dsd.Width - 1 )
				v01 = *(pSrcCurRow);
			else
				v01 = *(pSrcB0+4);

			if( x == 0 )
				vM1 = *(pSrcCurRow + (4 * (d3dsd.Width - 1)));
			else
				vM1 = *(pSrcB0-4);
			v10 = *(pSrcB1+0);
			v1M = *(pSrcB2+0);

			LONG iDu = (vM1-v01); // The delta-u bump value
			LONG iDv = (v1M-v10); // The delta-v bump value

			// The luminance bump value (land masses are less shiny)
			WORD uL = ( v00>1 ) ? 63 : 127;

			switch( BumpMapFormat )
			{
				case D3DFMT_V8U8:
					*pDstT++ = (BYTE)iDu;
					*pDstT++ = (BYTE)iDv;
					break;

				case D3DFMT_L6V5U5:
					*(WORD*)pDstT  = (WORD)( ( (iDu>>3) & 0x1f ) <<  0 );
					*(WORD*)pDstT |= (WORD)( ( (iDv>>3) & 0x1f ) <<  5 );
					*(WORD*)pDstT |= (WORD)( ( ( uL>>2) & 0x3f ) << 10 );
					pDstT += 2;
					break;

				case D3DFMT_X8L8V8U8:
					*pDstT++ = (BYTE)iDu;
					*pDstT++ = (BYTE)iDv;
					*pDstT++ = (BYTE)uL;
					*pDstT++ = (BYTE)0L;
					break;
			}

			// Move one pixel to the right (src is 32-bpp)
			pSrcB0+=4;
			pSrcB1+=4;
			pSrcB2+=4;
		}

		// Move to the next line
		pSrcCurRow += dwSrcPitch;
		pDstCurRow += dwDstPitch;
	}

	m_psBumpMap->UnlockRect(0);
	psBumpSrc->UnlockRect(0);

	return S_OK;
}

#endif