// ovY
// l^FDirectX9 Tv
#ifndef _BUMP_LENS_H_
#define _BUMP_LENS_H_

#include <d3dx9.h>
#include "macro.h"

struct BUMPVERTEX			// Y̒_tH[}bg`
{
	D3DXVECTOR3 v;			// W
	float		tu1,tv1;	// eNX`W1
	float		tu2,tv2;	// eNX`W2
};
#define FVF_BUMPVERTEX (D3DFVF_XYZ | D3DFVF_TEX2)

class CBumpLens
{
public:
	CBumpLens();
	~CBumpLens();
	void	Free();
	HRESULT Init(LPDIRECT3DDEVICE9 pD3DDev, UINT iWidth, UINT iHeight,
				LPDIRECT3DTEXTURE9 pTexBackground);
	HRESULT InitFromFile(LPDIRECT3DDEVICE9 pD3DDev, char *strTextureFilename,
						int power, LPDIRECT3DTEXTURE9 pTexBackground);
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev, float x, float y);
	
private:
	int 					m_iWidth,m_iHeight; // Y̑傫
	D3DXMATRIX				m_matView,m_matProj;
	LPDIRECT3DTEXTURE9		m_pTexBackground;	// wieNX`
	LPDIRECT3DTEXTURE9		m_psBumpMap;		// The actual bumpmap
	LPDIRECT3DVERTEXBUFFER9 m_pLensVB;			// Y̒_obt@
	
private:
	HRESULT CreateLensVB( LPDIRECT3DDEVICE9 pD3DDev );
	HRESULT CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev );
	HRESULT InitBumpMap( LPDIRECT3DDEVICE9 pD3DDev,
						LPDIRECT3DTEXTURE9 pTexLensBump, int power );
};

CBumpLens::CBumpLens()
{
	m_iWidth			= 0;
	m_iHeight			= 0;
	m_pTexBackground	= NULL;
	m_psBumpMap 		= NULL;
	m_pLensVB			= NULL;
	D3DXMatrixIdentity(&m_matView);
	D3DXMatrixIdentity(&m_matProj);
}

CBumpLens::~CBumpLens()
{
	Free();
}

void CBumpLens::Free()
{
	SAFE_RELEASE(m_psBumpMap);
	SAFE_RELEASE(m_pLensVB);
}

// 
// pD3DDev		  : [IN] IDirect3DDevice9 IuWFNg
// iWidth		  : [IN] Y̕
// iHeight		  : [IN] Y̍
// pTexBackground : [IN] wieNX`
HRESULT CBumpLens::Init(LPDIRECT3DDEVICE9 pD3DDev, UINT iWidth, UINT iHeight,
						LPDIRECT3DTEXTURE9 pTexBackground)
{
	if(!pD3DDev || !iWidth || !iHeight || !pTexBackground)
		return E_FAIL;
	
	Free();
	
	m_iWidth			= iWidth;
	m_iHeight			= iHeight;
	m_pTexBackground	= pTexBackground;
	
	HRESULT hr;
	hr = CreateLensVB(pD3DDev);
	hr = CreateBumpMap(pD3DDev);
	
	return hr;
}

// 摜t@C珉
// pD3DDev			   : [IN] IDirect3DDevice9 IuWFNg
// strTextureFilename  : [IN] ỸeNX`t@C
// power			   : [IN] Y̓x
// pTexBackground	   : [IN] wieNX`
HRESULT CBumpLens::InitFromFile(LPDIRECT3DDEVICE9 pD3DDev, char *strTextureFilename,
								int power, LPDIRECT3DTEXTURE9 pTexBackground)
{
	if(!pD3DDev || !strTextureFilename || !pTexBackground)
		return E_FAIL;
	
	Free();
	
	HRESULT 		hr;
	D3DXIMAGE_INFO	inf;
	
	hr = D3DXGetImageInfoFromFile(strTextureFilename, &inf);
	m_iWidth			= inf.Width;
	m_iHeight			= inf.Height;
	m_pTexBackground	= pTexBackground;
	
	hr = CreateLensVB(pD3DDev);
	
	LPDIRECT3DTEXTURE9	pTexLensBump;
	CreateTexture( pD3DDev, strTextureFilename, &pTexLensBump, D3DFMT_A8R8G8B8 );
	hr = InitBumpMap(pD3DDev, pTexLensBump, power);
	SAFE_RELEASE(pTexLensBump);
	
	return hr;
}

// _O
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
// x,y	   : [IN] YW
HRESULT CBumpLens::Render(LPDIRECT3DDEVICE9 pD3DDev, float x, float y)
{
	if(!pD3DDev || !m_pLensVB || !m_psBumpMap || !m_pTexBackground)
		return E_FAIL;
	
	HRESULT hr;
	DWORD	dwTex0Mag_pre,dwTex0Min_pre,dwTex1U_pre,dwTex1V_pre;
	DWORD	dwTex1Mag_pre,dwTex1Min_pre,dwTex1Mip_pre;
	DWORD	dwClr0Op_pre,dwClr0Arg1_pre,dwClr0Arg2_pre;
	DWORD	dwBump0Mat00_pre,dwBump0Mat01_pre,dwBump0Mat10_pre,dwBump0Mat11_pre;
	DWORD	dwBump0Scale_pre,dwBump0Offset_pre;
	DWORD	dwClr1Op_pre,dwClr1Arg1_pre,dwClr1Arg2_pre;
	DWORD	dwTex1Frag_pre,dwTex1Index_pre;
	DWORD	dwFVF_pre;
	D3DXMATRIX			matTex1_pre,matWorld_pre,matView_pre,matProj_pre;
	LPDIRECT3DTEXTURE9	pTex0_pre;
	LPDIRECT3DTEXTURE9	pTex1_pre;
	
	// ȑȌԂێ
	hr = pD3DDev->GetTexture( 0, (LPDIRECT3DBASETEXTURE9*) &pTex0_pre );
	hr = pD3DDev->GetTexture( 1, (LPDIRECT3DBASETEXTURE9*) &pTex1_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MAGFILTER, &dwTex0Mag_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MINFILTER, &dwTex0Min_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_ADDRESSU,  &dwTex1U_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_ADDRESSV,  &dwTex1V_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MAGFILTER, &dwTex1Mag_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MINFILTER, &dwTex1Min_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MIPFILTER, &dwTex1Mip_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLOROP, &dwClr0Op_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG1, &dwClr0Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG2, &dwClr0Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT00, &dwBump0Mat00_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT01, &dwBump0Mat01_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT10, &dwBump0Mat10_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT11, &dwBump0Mat11_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLSCALE, &dwBump0Scale_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, &dwBump0Offset_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLOROP, &dwClr1Op_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG1, &dwClr1Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG2, &dwClr1Arg2_pre );
	hr = pD3DDev->GetTransform( D3DTS_TEXTURE1, &matTex1_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, &dwTex1Frag_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, &dwTex1Index_pre);
	hr = pD3DDev->GetTransform( D3DTS_WORLD,	  &matWorld_pre );
	hr = pD3DDev->GetTransform( D3DTS_VIEW, 	  &matView_pre );
	hr = pD3DDev->GetTransform( D3DTS_PROJECTION, &matProj_pre );
	hr = pD3DDev->GetFVF( &dwFVF_pre );
	
	// eNX`ݒ
	hr = pD3DDev->SetTexture( 0, m_psBumpMap );
	hr = pD3DDev->SetTexture( 1, m_pTexBackground );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSU,  D3DTADDRESS_CLAMP );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSV,  D3DTADDRESS_CLAMP );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MIPFILTER, D3DTEXF_NONE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_BUMPENVMAP );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_CURRENT );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00, F2DW(0.2f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01, F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10, F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11, F2DW(0.2f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLSCALE, F2DW(1.0f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_CURRENT );
	// ϊsݒ
	hr = pD3DDev->SetTransform( D3DTS_VIEW, 	  &m_matView );
	hr = pD3DDev->SetTransform( D3DTS_PROJECTION, &m_matProj );
	
	// Generate texture coords depending on objects camera space position
	D3DXMATRIX mat;
	mat._11 = 0.501f; mat._12 = 0.0f;
	mat._21 = 0.0f;   mat._22 =-0.501f;
	mat._31 = 0.0f;   mat._32 = 0.0f;
	mat._41 = 0.501f; mat._42 = 0.501f;

	// Scale-by-z here
	D3DXVECTOR3 vEyePt( m_matView._41, m_matView._42, m_matView._43 );
	FLOAT		z = D3DXVec3Length( &vEyePt );
	mat._11 *= ( m_matProj._11 / ( m_matProj._33 * z + m_matProj._34 ) );
	mat._22 *= ( m_matProj._22 / ( m_matProj._33 * z + m_matProj._34 ) );

	hr = pD3DDev->SetTransform( D3DTS_TEXTURE1, &mat );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION | 1);

	// Position the lens
	D3DXMATRIX	matWorld;
	D3DXMatrixTranslation( &matWorld, x, y, 0.0f );
	hr = pD3DDev->SetTransform( D3DTS_WORLD, &matWorld );
	//	Render the lens
	hr = pD3DDev->SetFVF( FVF_BUMPVERTEX );
	hr = pD3DDev->SetStreamSource( 0, m_pLensVB, 0, sizeof(BUMPVERTEX) );
	hr = pD3DDev->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );

	// ɖ߂
	hr = pD3DDev->SetTexture( 0, pTex0_pre );
	hr = pD3DDev->SetTexture( 1, pTex1_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, dwTex0Mag_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, dwTex0Min_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSU,  dwTex1U_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_ADDRESSV,  dwTex1V_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, dwTex1Mag_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, dwTex1Min_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MIPFILTER, dwTex1Mip_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, dwClr0Op_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, dwClr0Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, dwClr0Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00, dwBump0Mat00_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01, dwBump0Mat01_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10, dwBump0Mat10_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11, dwBump0Mat11_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLSCALE, dwBump0Scale_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, dwBump0Offset_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP, dwClr1Op_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, dwClr1Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, dwClr1Arg2_pre );
	hr = pD3DDev->SetTransform( D3DTS_TEXTURE1, &matTex1_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, dwTex1Frag_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, dwTex1Index_pre);
	hr = pD3DDev->SetTransform( D3DTS_WORLD,	  &matWorld_pre );
	hr = pD3DDev->SetTransform( D3DTS_VIEW, 	  &matView_pre );
	hr = pD3DDev->SetTransform( D3DTS_PROJECTION, &matProj_pre );
	hr = pD3DDev->SetFVF( dwFVF_pre );
	
	return hr;
}

// Y̒_obt@쐬
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpLens::CreateLensVB( LPDIRECT3DDEVICE9 pD3DDev )
{
	HRESULT hr;
	SAFE_RELEASE(m_pLensVB);
	hr = pD3DDev->CreateVertexBuffer( 4*sizeof(BUMPVERTEX),D3DUSAGE_WRITEONLY,
									FVF_BUMPVERTEX, D3DPOOL_MANAGED, &m_pLensVB, NULL );
	if(SUCCEEDED(hr)){
		BUMPVERTEX* vLens;
		m_pLensVB->Lock( 0, 0, (void**)&vLens, 0 );
		vLens[0].v = D3DXVECTOR3( (float)-m_iWidth, (float)-m_iHeight, 0.0f );
		vLens[1].v = D3DXVECTOR3( (float)-m_iWidth, (float) m_iHeight, 0.0f );
		vLens[2].v = D3DXVECTOR3( (float) m_iWidth, (float)-m_iHeight, 0.0f );
		vLens[3].v = D3DXVECTOR3( (float) m_iWidth, (float) m_iHeight, 0.0f );
		vLens[0].tu1 = 0.0f; vLens[0].tv1 = 1.0f;
		vLens[1].tu1 = 0.0f; vLens[1].tv1 = 0.0f;
		vLens[2].tu1 = 1.0f; vLens[2].tv1 = 1.0f;
		vLens[3].tu1 = 1.0f; vLens[3].tv1 = 0.0f;
		m_pLensVB->Unlock();
	}
	// r[EvWFNVs쐬
	D3DSURFACE_DESC desc;
	LPDIRECT3DSURFACE9 pBackBuffer = NULL;
	pD3DDev->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
	pBackBuffer->GetDesc( &desc );
	pBackBuffer->Release();
	D3DXMatrixLookAtLH(&m_matView, &D3DXVECTOR3(0,0,-2001.0f),
						&D3DXVECTOR3(0,0,0), &D3DXVECTOR3(0,1,0));
	D3DXMatrixPerspectiveFovLH(&m_matProj, D3DX_PI/4,
						(float) desc.Width / desc.Height, 1.0f, 3000.0f);
	
	return hr;
}

// ỸeNX`쐬
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpLens::CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev )
{
	SAFE_RELEASE(m_psBumpMap);
	// Create the bumpmap's surface and texture objects
	if( FAILED( pD3DDev->CreateTexture( m_iWidth, m_iHeight, 1, 0,
		D3DFMT_V8U8, D3DPOOL_MANAGED, &m_psBumpMap, NULL ) ) )
	{
		return E_FAIL;
	}
	// Fill the bumpmap texels to simulate a lens
	D3DLOCKED_RECT d3dlr;
	m_psBumpMap->LockRect( 0, &d3dlr, 0, 0 );
	int x0,y0;
	int mid = m_iWidth / 2;
	for( y0 = 0; y0 < m_iHeight; y0++ )
	{
		CHAR* pDst = (CHAR*)d3dlr.pBits + y0*d3dlr.Pitch;
		for( x0 = 0; x0 < m_iWidth; x0++ )
		{
			DWORD x1 = ( (x0==m_iWidth-1)  ? x0 : x0+1 );
			DWORD y1 = ( (x0==m_iHeight-1) ? y0 : y0+1 );
			FLOAT fDistSq00 = (FLOAT)( (x0-mid)*(x0-mid) + (y0-mid)*(y0-mid) );
			FLOAT fDistSq01 = (FLOAT)( (x1-mid)*(x1-mid) + (y0-mid)*(y0-mid) );
			FLOAT fDistSq10 = (FLOAT)( (x0-mid)*(x0-mid) + (y1-mid)*(y1-mid) );
			FLOAT v00 = ( fDistSq00 > (mid*mid) ) ? 0.0f : sqrtf( (mid*mid) - fDistSq00 );
			FLOAT v01 = ( fDistSq01 > (mid*mid) ) ? 0.0f : sqrtf( (mid*mid) - fDistSq01 );
			FLOAT v10 = ( fDistSq10 > (mid*mid) ) ? 0.0f : sqrtf( (mid*mid) - fDistSq10 );
			FLOAT iDu = (128/D3DX_PI)*atanf(v00-v01); // The delta-u bump value
			FLOAT iDv = (128/D3DX_PI)*atanf(v00-v10); // The delta-v bump value
			*pDst++ = (CHAR)(iDu);
			*pDst++ = (CHAR)(iDv);
		}
	}
	m_psBumpMap->UnlockRect(0);

	return S_OK;
}

// ov}bv
// pD3DDev		: [IN] IDirect3DDevice9 IuWFNg
// pTexLensBump : [IN] Y̌eNX`
// power		: [IN] Y̓x
HRESULT CBumpLens::InitBumpMap( LPDIRECT3DDEVICE9 pD3DDev,
								LPDIRECT3DTEXTURE9 pTexLensBump, int power )
{
	LPDIRECT3DTEXTURE9	psBumpSrc = pTexLensBump;
	D3DSURFACE_DESC 	d3dsd;
	D3DLOCKED_RECT		d3dlr;
	D3DFORMAT			BumpMapFormat = D3DFMT_V8U8;

	SAFE_RELEASE(m_psBumpMap);
	psBumpSrc->GetLevelDesc( 0, &d3dsd );
	// Create the bumpmap's surface and texture objects
	if( FAILED( pD3DDev->CreateTexture( d3dsd.Width, d3dsd.Height, 1, 0,
		BumpMapFormat, D3DPOOL_MANAGED, &m_psBumpMap, NULL ) ) )
	{
		return E_FAIL;
	}

	// Fill the bits of the new texture surface with bits from
	// a private format.
	psBumpSrc->LockRect( 0, &d3dlr, 0, 0 );
	DWORD dwSrcPitch = (DWORD)d3dlr.Pitch;
	BYTE* pSrcTopRow = (BYTE*)d3dlr.pBits;
	BYTE* pSrcCurRow = pSrcTopRow;
	BYTE* pSrcBotRow = pSrcTopRow + (dwSrcPitch * (d3dsd.Height - 1) );

	m_psBumpMap->LockRect( 0, &d3dlr, 0, 0 );
	DWORD dwDstPitch = (DWORD)d3dlr.Pitch;
	BYTE* pDstTopRow = (BYTE*)d3dlr.pBits;
	BYTE* pDstCurRow = pDstTopRow;

	for( DWORD y=0; y<d3dsd.Height; y++ )
	{
		BYTE* pSrcB0; // addr of current pixel
		BYTE* pSrcB1; // addr of pixel below current pixel, wrapping to top if necessary
		BYTE* pSrcB2; // addr of pixel above current pixel, wrapping to bottom if necessary
		BYTE* pDstT;  // addr of dest pixel;

		pSrcB0 = pSrcCurRow;

		if( y == d3dsd.Height - 1)
			pSrcB1 = pSrcTopRow;
		else
			pSrcB1 = pSrcCurRow + dwSrcPitch;

		if( y == 0 )
			pSrcB2 = pSrcBotRow;
		else
			pSrcB2 = pSrcCurRow - dwSrcPitch;

		pDstT = pDstCurRow;

		for( DWORD x=0; x<d3dsd.Width; x++ )
		{
			LONG v00; // Current pixel
			LONG v01; // Pixel to the right of current pixel, wrapping to left edge if necessary
			LONG vM1; // Pixel to the left of current pixel, wrapping to right edge if necessary
			LONG v10; // Pixel one line below.
			LONG v1M; // Pixel one line above.

			v00 = *(pSrcB0+0);

			if( x == d3dsd.Width - 1 )
				v01 = *(pSrcCurRow);
			else
				v01 = *(pSrcB0+4);

			if( x == 0 )
				vM1 = *(pSrcCurRow + (4 * (d3dsd.Width - 1)));
			else
				vM1 = *(pSrcB0-4);
			v10 = *(pSrcB1+0);
			v1M = *(pSrcB2+0);

			LONG iDu = (vM1-v01)*power; // The delta-u bump value
			LONG iDv = (v1M-v10)*power; // The delta-v bump value

			// The luminance bump value (land masses are less shiny)
			WORD uL = ( v00>1 ) ? 63 : 127;

			switch( BumpMapFormat )
			{
				case D3DFMT_V8U8:
					*pDstT++ = (BYTE)iDu;
					*pDstT++ = (BYTE)iDv;
					break;

				case D3DFMT_L6V5U5:
					*(WORD*)pDstT  = (WORD)( ( (iDu>>3) & 0x1f ) <<  0 );
					*(WORD*)pDstT |= (WORD)( ( (iDv>>3) & 0x1f ) <<  5 );
					*(WORD*)pDstT |= (WORD)( ( ( uL>>2) & 0x3f ) << 10 );
					pDstT += 2;
					break;

				case D3DFMT_X8L8V8U8:
					*pDstT++ = (BYTE)iDu;
					*pDstT++ = (BYTE)iDv;
					*pDstT++ = (BYTE)uL;
					*pDstT++ = (BYTE)0L;
					break;
			}

			// Move one pixel to the right (src is 32-bpp)
			pSrcB0+=4;
			pSrcB1+=4;
			pSrcB2+=4;
		}

		// Move to the next line
		pSrcCurRow += dwSrcPitch;
		pDstCurRow += dwDstPitch;
	}

	m_psBumpMap->UnlockRect(0);
	psBumpSrc->UnlockRect(0);

	return S_OK;
}

#endif