// ov
// l^FDirectX9 Tv
#ifndef _BUMP_UNDERWATER_H_
#define _BUMP_UNDERWATER_H_

#include <d3dx9.h>
#include "macro.h"

struct BUMPVERTEX2			// ̒_tH[}bg`
{
	D3DXVECTOR3 v;			// W
	D3DXVECTOR3 n;			// @xNg
	float		tu1,tv1;	// eNX`W1
	float		tu2,tv2;	// eNX`W2
};
#define FVF_BUMPVERTEX2 (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX2)

class CBumpUnderWater
{
public:
	CBumpUnderWater();
	~CBumpUnderWater();
	void	Free();
	HRESULT Init(LPDIRECT3DDEVICE9 pD3DDev, float fWidth, float fHeight,
				float fFrequency, LPDIRECT3DTEXTURE9 pTex);
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev, float fSpeed, float fAmplitude);
	
private:
	float					m_fWidth,m_fHeight; // ̑傫
	float					m_fFrequency;		// U
	LPDIRECT3DTEXTURE9		m_pTex; 			// eNX`
	LPDIRECT3DTEXTURE9		m_pBumpMap; 		// The actual bumpmap
	LPDIRECT3DVERTEXBUFFER9 m_pWaterVB; 		// ̒_obt@
	
private:
	HRESULT CreateWaterVB( LPDIRECT3DDEVICE9 pD3DDev );
	HRESULT CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev );
};

CBumpUnderWater::CBumpUnderWater()
{
	m_fWidth	= 0;
	m_fHeight	= 0;
	m_pTex		= NULL;
	m_pBumpMap	= NULL;
	m_pWaterVB	= NULL;
}

CBumpUnderWater::~CBumpUnderWater()
{
	Free();
}

void CBumpUnderWater::Free()
{
	SAFE_RELEASE(m_pBumpMap);
	SAFE_RELEASE(m_pWaterVB);
}

// 
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
// fWidth  : [IN] ̕
// fHeight : [IN] ̍
// pTex    : [IN] eNX`
HRESULT CBumpUnderWater::Init(LPDIRECT3DDEVICE9 pD3DDev, float fWidth, float fHeight,
								float fFrequency, LPDIRECT3DTEXTURE9 pTex)
{
	if(!pD3DDev || !fWidth || !fHeight || !pTex)
		return E_FAIL;
	
	Free();
	
	m_fWidth		= fWidth;
	m_fHeight		= fHeight;
	m_fFrequency	= fFrequency;
	m_pTex			= pTex;
	
	HRESULT hr;
	hr = CreateWaterVB(pD3DDev);
	hr = CreateBumpMap(pD3DDev);
	
	return hr;
}

// _O
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
// fSpeed  : [IN] x
HRESULT CBumpUnderWater::Render(LPDIRECT3DDEVICE9 pD3DDev, float fSpeed, float fAmplitude)
{
	if(!pD3DDev || !m_pWaterVB || !m_pBumpMap || !m_pTex)
		return E_FAIL;
	
	HRESULT hr;
	DWORD	dwTex0Mag_pre,dwTex0Min_pre;
	DWORD	dwTex1Mag_pre,dwTex1Min_pre;
	DWORD	dwClr0Op_pre,dwClr0Arg1_pre,dwClr0Arg2_pre,dwAlp0Op_pre;
	DWORD	dwBump0Mat00_pre,dwBump0Mat01_pre,dwBump0Mat10_pre,dwBump0Mat11_pre;
	DWORD	dwClr1Op_pre,dwClr1Arg1_pre,dwClr1Arg2_pre,dwAlp1Op_pre;
	DWORD	dwClr2Op_pre,dwAlp2Op_pre;
	DWORD	dwTex0Index_pre,dwTex1Index_pre;
	DWORD	dwDither_pre,dwFVF_pre;
	D3DXMATRIX			matView_pre,matProj_pre;
	LPDIRECT3DTEXTURE9	pTex0_pre;
	LPDIRECT3DTEXTURE9	pTex1_pre;
	
	// ȑȌԂێ
	hr = pD3DDev->GetRenderState( D3DRS_DITHERENABLE, &dwDither_pre );
	hr = pD3DDev->GetTexture( 0, (LPDIRECT3DBASETEXTURE9*) &pTex0_pre );
	hr = pD3DDev->GetTexture( 1, (LPDIRECT3DBASETEXTURE9*) &pTex1_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MAGFILTER, &dwTex0Mag_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MINFILTER, &dwTex0Min_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MAGFILTER, &dwTex1Mag_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MINFILTER, &dwTex1Min_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLOROP, &dwClr0Op_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG1, &dwClr0Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG2, &dwClr0Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_ALPHAOP, &dwAlp0Op_pre);
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, &dwTex0Index_pre);
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT00, &dwBump0Mat00_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT01, &dwBump0Mat01_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT10, &dwBump0Mat10_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT11, &dwBump0Mat11_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLOROP, &dwClr1Op_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG1, &dwClr1Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG2, &dwClr1Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_ALPHAOP, &dwAlp1Op_pre);
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, &dwTex1Index_pre);
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_COLOROP, &dwClr2Op_pre );
	hr = pD3DDev->GetTextureStageState( 2, D3DTSS_ALPHAOP, &dwAlp2Op_pre);
	hr = pD3DDev->GetFVF( &dwFVF_pre );
	
	// Render the water
	
	static FLOAT	fTime = 0;
	BUMPVERTEX2*	vWaterVertices;
	m_pWaterVB->Lock( 0, 0, (void**)&vWaterVertices, 0 );
	vWaterVertices[0].tu1 =  0.000f; vWaterVertices[0].tv1 = 0.5f * fTime + 2.0f;
	vWaterVertices[1].tu1 =  0.000f; vWaterVertices[1].tv1 = 0.5f * fTime;
	vWaterVertices[2].tu1 =  1.000f; vWaterVertices[2].tv1 = 0.5f * fTime;
	vWaterVertices[3].tu1 =  1.000f; vWaterVertices[3].tv1 = 0.5f * fTime + 2.0f;
	m_pWaterVB->Unlock();
	fTime += fSpeed;

	hr = pD3DDev->SetTexture( 0, m_pBumpMap );
	hr = pD3DDev->SetTexture( 1, m_pTex );

	hr = pD3DDev->SetRenderState( D3DRS_DITHERENABLE, TRUE );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00, F2DW(fAmplitude) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01, F2DW(0.00f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10, F2DW(0.00f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11, F2DW(fAmplitude) );

	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0 );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP,	 D3DTOP_BUMPENVMAP );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_CURRENT );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP,	 D3DTOP_DISABLE );

	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, 1 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,	 D3DTOP_SELECTARG1 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_CURRENT );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,	 D3DTOP_DISABLE );

	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLOROP,	 D3DTOP_DISABLE );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_ALPHAOP,	 D3DTOP_DISABLE );

	hr = pD3DDev->SetFVF( FVF_BUMPVERTEX2 );
	hr = pD3DDev->SetStreamSource( 0, m_pWaterVB, 0, sizeof(BUMPVERTEX2) );
	hr = pD3DDev->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );

	// ɖ߂
	hr = pD3DDev->SetRenderState( D3DRS_DITHERENABLE, dwDither_pre );
	hr = pD3DDev->SetTexture( 0, pTex0_pre );
	hr = pD3DDev->SetTexture( 1, pTex1_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, dwTex0Mag_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, dwTex0Min_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, dwTex1Mag_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, dwTex1Min_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, dwClr0Op_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, dwClr0Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, dwClr0Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, dwAlp0Op_pre);
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, dwTex0Index_pre);
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00, dwBump0Mat00_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01, dwBump0Mat01_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10, dwBump0Mat10_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11, dwBump0Mat11_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP, dwClr1Op_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, dwClr1Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG2, dwClr1Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP, dwAlp1Op_pre);
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, dwTex1Index_pre);
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_COLOROP, dwClr2Op_pre );
	hr = pD3DDev->SetTextureStageState( 2, D3DTSS_ALPHAOP, dwAlp2Op_pre);
	hr = pD3DDev->SetFVF( dwFVF_pre );
	
	return hr;
}

// ̒_obt@쐬
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpUnderWater::CreateWaterVB( LPDIRECT3DDEVICE9 pD3DDev )
{
	HRESULT hr;
	SAFE_RELEASE(m_pWaterVB);
	hr = pD3DDev->CreateVertexBuffer( 4*sizeof(BUMPVERTEX2),D3DUSAGE_WRITEONLY,
									FVF_BUMPVERTEX2, D3DPOOL_MANAGED, &m_pWaterVB, NULL );
	if(SUCCEEDED(hr)){
		BUMPVERTEX2* vWater;
		m_pWaterVB->Lock( 0, 0, (void**)&vWater, 0 );
		vWater[0].v = D3DXVECTOR3( (float)-m_fWidth, (float)-m_fHeight, 0.0f );
		vWater[1].v = D3DXVECTOR3( (float)-m_fWidth, (float) m_fHeight, 0.0f );
		vWater[2].v = D3DXVECTOR3( (float) m_fWidth, (float)-m_fHeight, 0.0f );
		vWater[3].v = D3DXVECTOR3( (float) m_fWidth, (float) m_fHeight, 0.0f );
		vWater[0].n = D3DXVECTOR3( 0, 1, 0 );
		vWater[1].n = D3DXVECTOR3( 0, 1, 0 );
		vWater[2].n = D3DXVECTOR3( 0, 1, 0 );
		vWater[3].n = D3DXVECTOR3( 0, 1, 0 );
		vWater[0].tu2 = 0.0f; vWater[0].tv2 = 1.0f;
		vWater[1].tu2 = 0.0f; vWater[1].tv2 = 0.0f;
		vWater[2].tu2 = 1.0f; vWater[2].tv2 = 1.0f;
		vWater[3].tu2 = 1.0f; vWater[3].tv2 = 0.0f;
		m_pWaterVB->Unlock();
	}
	return hr;
}

// ̃eNX`쐬
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpUnderWater::CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev )
{
	SAFE_RELEASE(m_pBumpMap);

	DWORD iWidth  = 256;
	DWORD iHeight = 256;

	// Create the bumpmap's surface and texture objects
	if( FAILED( pD3DDev->CreateTexture( iWidth, iHeight, 1, 0,
		D3DFMT_V8U8, D3DPOOL_MANAGED, &m_pBumpMap, NULL ) ) )
	{
		return E_FAIL;
	}

	// Fill the bumpmap texels to simulate a lens
	D3DLOCKED_RECT lrDst;
	m_pBumpMap->LockRect( 0, &lrDst, 0, 0 );
	BYTE* pDst		 = (BYTE*)lrDst.pBits;

	for( DWORD y=0; y<iHeight; y++ )
	{
		for( DWORD x=0; x<iWidth; x++ )
		{
			FLOAT fx = x/(FLOAT)iWidth	- 0.5f;
			FLOAT fy = y/(FLOAT)iHeight - 0.5f;

			CHAR iDu = (CHAR)(64 * cosf(m_fFrequency * (fx + fy) * D3DX_PI));
			CHAR iDv = (CHAR)(64 * sinf(m_fFrequency * (fx + fy) * D3DX_PI));

			pDst[2 * x + 0] = iDu;
			pDst[2 * x + 1] = iDv;
		}
		pDst += lrDst.Pitch;
	}

	m_pBumpMap->UnlockRect(0);

	return S_OK;
}

#endif