// gov
// l^FDirectX9 Tv
#ifndef _BUMP_WAVES_H_
#define _BUMP_WAVES_H_

#include <d3dx9.h>
#include "macro.h"
#include "fvf.h"

class CBumpWaves
{
public:
	CBumpWaves();
	~CBumpWaves();
	void	Free();
	HRESULT Init(LPDIRECT3DDEVICE9 pD3DDev, DWORD dwWidth, DWORD dwHeight,
				D3DFORMAT d3dBumpFormat, LPDIRECT3DTEXTURE9 pTexBackground);
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev, float fSpeed);
	void	ChangeBumpMap();
	int 	GetFlagWave() { return m_flagWave; }
	void	SetFlagWave(int flagWave) { m_flagWave = flagWave; }
	
private:
	D3DCAPS9				m_d3dCaps;
	UINT					m_nTriangles;		// Number of triangles in the ground grid
	int 					m_flagWave; 		// gtOi0:1A1:2A2:0+1A3:QA4:2+3j
	LPDIRECT3DTEXTURE9		m_pTexBackground;	// wieNX`
	LPDIRECT3DTEXTURE9		m_pBumpMap; 		// The actual bumpmap
	LPDIRECT3DVERTEXBUFFER9 m_pWaterVB; 		// g̒_obt@
	
private:
	HRESULT CreateWaterVB( LPDIRECT3DDEVICE9 pD3DDev );
	LPDIRECT3DTEXTURE9 CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev,
			DWORD dwWidth, DWORD dwHeight, D3DFORMAT d3dBumpFormat );
};

CBumpWaves::CBumpWaves()
{
	ZeroMemory(&m_d3dCaps, sizeof(m_d3dCaps));
	m_nTriangles		= 0;
	m_flagWave			= 4;
	m_pTexBackground	= NULL;
	m_pBumpMap			= NULL;
	m_pWaterVB			= NULL;
}

CBumpWaves::~CBumpWaves()
{
	Free();
}

void CBumpWaves::Free()
{
	SAFE_RELEASE(m_pBumpMap);
	SAFE_RELEASE(m_pWaterVB);
}

// 
// pD3DDev		  : [IN] IDirect3DDevice9 IuWFNg
// iWidth		  : [IN] geNX`̕
// iHeight		  : [IN] geNX`̍
// d3dBumpFormat  : [IN] geNX`̃tH[}bg
// pTexBackground : [IN] wieNX`
HRESULT CBumpWaves::Init(LPDIRECT3DDEVICE9 pD3DDev, DWORD dwWidth, DWORD dwHeight,
						D3DFORMAT d3dBumpFormat, LPDIRECT3DTEXTURE9 pTexBackground)
{
	if(!pD3DDev || !dwWidth || !dwHeight || !pTexBackground)
		return E_FAIL;
	
	Free();
	
	m_pTexBackground = pTexBackground;
	
	HRESULT hr;
	
	hr = pD3DDev->GetDeviceCaps(&m_d3dCaps);
	hr = CreateWaterVB(pD3DDev);
	m_pBumpMap = CreateBumpMap(pD3DDev, dwWidth, dwHeight, d3dBumpFormat);
	if(m_pBumpMap){
		ChangeBumpMap();
	}
	
	return hr;
}

// _O
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
// fSpeed  : [IN] g̑x
HRESULT CBumpWaves::Render(LPDIRECT3DDEVICE9 pD3DDev, float fSpeed)
{
	if(!pD3DDev || !m_pWaterVB || !m_pBumpMap || !m_pTexBackground)
		return E_FAIL;
	
	HRESULT hr;
	DWORD	dwTex0Mag_pre,dwTex0Min_pre,dwTex0U_pre,dwTex0V_pre;
	DWORD	dwTex1Mag_pre,dwTex1Min_pre;
	DWORD	dwClr0Op_pre,dwClr0Arg1_pre,dwClr0Arg2_pre;
	DWORD	dwBump0Mat00_pre,dwBump0Mat01_pre,dwBump0Mat10_pre,dwBump0Mat11_pre;
	DWORD	dwBump0Scale_pre,dwBump0Offset_pre;
	DWORD	dwClr1Op_pre,dwClr1Arg1_pre;
	DWORD	dwTex1Frag_pre,dwTex1Index_pre;
	DWORD	dwDither_pre,dwFVF_pre;
	D3DXMATRIX			matTex1;
	LPDIRECT3DTEXTURE9	pTex0_pre;
	LPDIRECT3DTEXTURE9	pTex1_pre;
	
	// ȑȌԂێ
	hr = pD3DDev->GetRenderState( D3DRS_DITHERENABLE, &dwDither_pre );
	hr = pD3DDev->GetTexture( 0, (LPDIRECT3DBASETEXTURE9*) &pTex0_pre );
	hr = pD3DDev->GetTexture( 1, (LPDIRECT3DBASETEXTURE9*) &pTex1_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MAGFILTER, &dwTex0Mag_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_MINFILTER, &dwTex0Min_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_ADDRESSU,  &dwTex0U_pre );
	hr = pD3DDev->GetSamplerState( 0, D3DSAMP_ADDRESSV,  &dwTex0V_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MAGFILTER, &dwTex1Mag_pre );
	hr = pD3DDev->GetSamplerState( 1, D3DSAMP_MINFILTER, &dwTex1Min_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLOROP, &dwClr0Op_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG1, &dwClr0Arg1_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_COLORARG2, &dwClr0Arg2_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT00, &dwBump0Mat00_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT01, &dwBump0Mat01_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT10, &dwBump0Mat10_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVMAT11, &dwBump0Mat11_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLSCALE, &dwBump0Scale_pre );
	hr = pD3DDev->GetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, &dwBump0Offset_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLOROP, &dwClr1Op_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_COLORARG1, &dwClr1Arg1_pre );
	hr = pD3DDev->GetTransform( D3DTS_TEXTURE1, &matTex1 );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, &dwTex1Frag_pre );
	hr = pD3DDev->GetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, &dwTex1Index_pre);
	hr = pD3DDev->GetFVF( &dwFVF_pre );
	
	// Setup the bump matrix
	// Min r is 0.04 if amplitude is 32 to miss temporal aliasing
	// Max r is 0.16 for amplitude is 8 to miss spatial aliasing
	static FLOAT	fTime = 0;
	FLOAT			r = 0.04f;
	float			fMat11,fMat12,fMat21,fMat22;
	fMat11 =  r * cosf( fTime );
	fMat12 = -r * sinf( fTime );
	fMat21 =  r * sinf( fTime );
	fMat22 =  r * cosf( fTime );
	fTime += fSpeed;
	
	// fBUL
	hr = pD3DDev->SetRenderState( D3DRS_DITHERENABLE, TRUE );
	
	// Set up texture stage 0's states for the bumpmap
	hr = pD3DDev->SetTexture( 0, m_pBumpMap );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSU,   D3DTADDRESS_CLAMP );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSV,   D3DTADDRESS_CLAMP );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00,   F2DW( fMat11 ) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01,   F2DW( fMat12 ) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10,   F2DW( fMat21 ) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11,   F2DW( fMat22 ) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLSCALE,  F2DW(0.8f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, F2DW(0.0f) );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP,	D3DTOP_BUMPENVMAPLUMINANCE );

	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_CURRENT );

	// Set up texture stage 1's states for the environment map
	hr = pD3DDev->SetTexture( 1, m_pTexBackground );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, D3DTEXF_LINEAR );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,	D3DTOP_SELECTARG1 );

	// Set up projected texture coordinates
	// tu = (0.8x + 0.5z) / z
	// tv = (0.8y - 0.5z) / z
	D3DXMATRIX mat;
	mat._11 = 0.91f; mat._12 = 0.0f; mat._13 = 0.0f;
	mat._21 = 0.0f; mat._22 = 0.91f; mat._23 = 0.0f;
	mat._31 = 0.5f; mat._32 =-0.5f; mat._33 = 1.0f;
	mat._41 = 0.0f; mat._42 = 0.0f; mat._43 = 0.0f;
	
	hr = pD3DDev->SetTransform( D3DTS_TEXTURE1, &mat );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT3|D3DTTFF_PROJECTED );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION | 1 );
	
	// Render the water
	hr = pD3DDev->SetFVF( FVF_VERTEX_T );
	hr = pD3DDev->SetStreamSource( 0, m_pWaterVB, 0, sizeof(VERTEX_T) );
	hr = pD3DDev->DrawPrimitive( D3DPT_TRIANGLELIST, 0, m_nTriangles );

	// ɖ߂
	hr = pD3DDev->SetRenderState( D3DRS_DITHERENABLE, dwDither_pre );
	hr = pD3DDev->SetTexture( 0, pTex0_pre );
	hr = pD3DDev->SetTexture( 1, pTex1_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MAGFILTER, dwTex0Mag_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_MINFILTER, dwTex0Min_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSU,  dwTex0U_pre );
	hr = pD3DDev->SetSamplerState( 0, D3DSAMP_ADDRESSV,  dwTex0V_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MAGFILTER, dwTex1Mag_pre );
	hr = pD3DDev->SetSamplerState( 1, D3DSAMP_MINFILTER, dwTex1Min_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, dwClr0Op_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, dwClr0Arg1_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, dwClr0Arg2_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT00, dwBump0Mat00_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT01, dwBump0Mat01_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT10, dwBump0Mat10_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVMAT11, dwBump0Mat11_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLSCALE, dwBump0Scale_pre );
	hr = pD3DDev->SetTextureStageState( 0, D3DTSS_BUMPENVLOFFSET, dwBump0Offset_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP, dwClr1Op_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_COLORARG1, dwClr1Arg1_pre );
	hr = pD3DDev->SetTransform( D3DTS_TEXTURE1, &matTex1 );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, dwTex1Frag_pre );
	hr = pD3DDev->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, dwTex1Index_pre);
	hr = pD3DDev->SetFVF( dwFVF_pre );
	
	return hr;
}

// g̒_obt@쐬
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CBumpWaves::CreateWaterVB( LPDIRECT3DDEVICE9 pD3DDev )
{
	HRESULT hr;
	SAFE_RELEASE(m_pWaterVB);
	
	UINT	n;				// Number of vertices in the ground grid along X
	UINT	m;				// Number of vertices in the ground grid along Z
	if( m_d3dCaps.TextureCaps & D3DPTEXTURECAPS_PROJECTED )
	{
		n = 2;			   // Number of vertices in the ground grid along X
		m = 2;			   // Number of vertices in the ground grid along Z
	}
	else
	{
		n = 8;			   // Number of vertices in the ground grid along X
		m = 8;			   // Number of vertices in the ground grid along Z
	}
	m_nTriangles = (n-1)*(m-1)*2;	// Number of triangles in the ground
	
	hr = pD3DDev->CreateVertexBuffer( m_nTriangles*3*sizeof(VERTEX_T),
									D3DUSAGE_WRITEONLY, FVF_VERTEX_T,
									D3DPOOL_MANAGED, &m_pWaterVB, NULL );
	if(SUCCEEDED(hr)){
		VERTEX_T* v;
		m_pWaterVB->Lock( 0, 0, (void**)&v, 0 );
		float dX = 20.0f/(n-1);
		float dZ = 12.5f/(m-1);
		float x0 = -10.0f;
		float z0 = -12.5f;
		float dU = 1.0f/(n-1);
		float dV = 0.7f/(m-1);
		UINT x,z,k = 0;
		for (z=0; z < (m-1); z++)
		{
			for (x=0; x < (n-1); x++)
			{
				v[k].v	= D3DXVECTOR3(x0 + x*dX, 0.0f, z0 + z*dZ );
				v[k].tu = x*dU;
				v[k].tv = z*dV;
				k++;
				v[k].v	= D3DXVECTOR3(x0 + x*dX, 0.0f, z0 + (z+1)*dZ );
				v[k].tu = x*dU;
				v[k].tv = (z+1)*dV;
				k++;
				v[k].v	= D3DXVECTOR3(x0 + (x+1)*dX, 0.0f, z0 + (z+1)*dZ );
				v[k].tu = (x+1)*dU;
				v[k].tv = (z+1)*dV;
				k++;
				v[k].v	= D3DXVECTOR3(x0 + x*dX, 0.0f, z0 + z*dZ );
				v[k].tu = x*dU;
				v[k].tv = z*dV;
				k++;
				v[k].v	= D3DXVECTOR3(x0 + (x+1)*dX, 0.0f, z0 + (z+1)*dZ );
				v[k].tu = (x+1)*dU;
				v[k].tv = (z+1)*dV;
				k++;
				v[k].v	= D3DXVECTOR3(x0 + (x+1)*dX, 0.0f, z0 + z*dZ );
				v[k].tu = (x+1)*dU;
				v[k].tv = z*dV;
				k++;
			}
		}
		m_pWaterVB->Unlock();
	}
	
	return hr;
}

// oveNX`쐬
LPDIRECT3DTEXTURE9 CBumpWaves::CreateBumpMap( LPDIRECT3DDEVICE9 pD3DDev,
					DWORD dwWidth, DWORD dwHeight, D3DFORMAT d3dBumpFormat )
{
	if(!pD3DDev)
		return NULL;
	
	LPDIRECT3DTEXTURE9	psBumpMap;

	// Check if the device can create the format
	LPDIRECT3D9 	pD3D = NULL;
	D3DDISPLAYMODE	mode;
	pD3DDev->GetDirect3D(&pD3D);
	pD3D->GetAdapterDisplayMode( m_d3dCaps.AdapterOrdinal, &mode );
	if( FAILED( pD3D->CheckDeviceFormat( m_d3dCaps.AdapterOrdinal,
										m_d3dCaps.DeviceType,
										mode.Format,
										0, D3DRTYPE_TEXTURE, d3dBumpFormat ) ) )
		return NULL;

	// Create the bump map texture
	if( FAILED( pD3DDev->CreateTexture( dwWidth, dwHeight, 1, 0 /* Usage */,
										d3dBumpFormat, D3DPOOL_MANAGED,
										&psBumpMap, NULL ) ) )
		return NULL;

	return psBumpMap;
}

// g̕ύX
void CBumpWaves::ChangeBumpMap()
{ 
	// Lock the surface and write in some bumps for the waves
	D3DLOCKED_RECT	d3dlr;
	D3DSURFACE_DESC desc;
	m_pBumpMap->LockRect( 0, &d3dlr, 0, 0 );
	m_pBumpMap->GetLevelDesc( 0, &desc );
	CHAR* pDst = (CHAR*)d3dlr.pBits;
	CHAR  iDu, iDv;

	for( DWORD y=0; y<desc.Height; y++ )
	{
		CHAR* pPixel = pDst;

		for( DWORD x=0; x<desc.Width; x++ )
		{
			FLOAT fx = x/(FLOAT)desc.Width - 0.5f;
			FLOAT fy = y/(FLOAT)desc.Height - 0.2f;

			FLOAT r = sqrtf( fx*fx + fy*fy );

			switch(m_flagWave)
			{
			case 0: 	// 1
				iDu = (CHAR)( 32 * cosf( 150.0f * ( fx + fy ) ) );
				iDv = (CHAR)( 32 * sinf( 150.0f * ( fx + fy ) ) );
				break;
			case 1: 	// 2
				iDu = (CHAR)( 16 * cosf( 140.0f * ( fx * 0.85f - fy ) ) );
				iDv = (CHAR)( 16 * sinf( 140.0f * ( fx * 0.85f - fy ) ) );
				break;
			case 2: 	// 1 + 2
				iDu  = (CHAR)( 32 * cosf( 150.0f * ( fx + fy ) ) );
				iDu += (CHAR)( 16 * cosf( 140.0f * ( fx * 0.85f - fy ) ) );
				iDv  = (CHAR)( 32 * sinf( 150.0f * ( fx + fy ) ) );
				iDv += (CHAR)( 16 * sinf( 140.0f * ( fx * 0.85f - fy ) ) );
				break;
			case 3: 	// Q
				iDu = (CHAR)( 64 * cosf( 300.0f * r ) * expf( -r * 5.0f ) );
				iDv = (CHAR)( 64 * sinf( 300.0f * r ) * expf( -r * 5.0f ) );
				break;
			case 4: 	// 1 + 2 + Q
				iDu  = (CHAR)( 64 * cosf( 300.0f * r ) * expf( -r * 5.0f ) );
				iDu += (CHAR)( 32 * cosf( 150.0f * ( fx + fy ) ) );
				iDu += (CHAR)( 16 * cosf( 140.0f * ( fx * 0.85f - fy ) ) );
				iDv  = (CHAR)( 64 * sinf( 300.0f * r ) * expf( -r * 5.0f ) );
				iDv += (CHAR)( 32 * sinf( 150.0f * ( fx + fy ) ) );
				iDv += (CHAR)( 16 * sinf( 140.0f * ( fx * 0.85f - fy ) ) );
				break;
			}
			*pPixel++ = iDu;
			*pPixel++ = iDv;
		}
		pDst += d3dlr.Pitch;
	}
	m_pBumpMap->UnlockRect(0);
}// ChangeBumpMap

#endif