// L[u}bsO̒ۃNX
// l^FDirectX9 Tv
#ifndef _CUBEMAP_H_
#define _CUBEMAP_H_

#include <d3dx9.h>
#include "macro.h"

class CCubeMap
{
public:
	CCubeMap();
	virtual ~CCubeMap();
	void	Free();
	HRESULT Init( LPDIRECT3DDEVICE9 pD3DDev, UINT iCubemapSize = 256 );
	HRESULT RenderSceneIntoCubeMap( LPDIRECT3DDEVICE9 pD3DDev, D3DXVECTOR3 v3EyePt,
							D3DXMATRIX matWorld, D3DXMATRIX matT, D3DXMATRIX matRY );
	virtual HRESULT RenderScene( LPDIRECT3DDEVICE9 pD3DDev, D3DXMATRIX matWorld,
								D3DXMATRIX matT, D3DXMATRIX matRY ) = 0;
	LPDIRECT3DCUBETEXTURE9	GetTexture() { return m_pCubeMap; }
	
private:
	D3DCAPS9				m_d3dCaps;			// foCX
	D3DSURFACE_DESC 		m_d3dsdBackBuffer;	// obNobt@
	LPDIRECT3DCUBETEXTURE9	m_pCubeMap; 		// L[u}bv eNX`
	LPDIRECT3DSURFACE9		m_pZBuffer; 		// L[u}bvZobt@
	
private:
	D3DXMATRIX GetCubeMapViewMatrix( DWORD dwFace, D3DXVECTOR3 v3EyePt,
							D3DXMATRIX matWorld, D3DXMATRIX matT, D3DXMATRIX matRY );
};

CCubeMap::CCubeMap()
{
	ZeroMemory(&m_d3dCaps, sizeof(m_d3dCaps));
	ZeroMemory(&m_d3dsdBackBuffer, sizeof(m_d3dsdBackBuffer));
	m_pCubeMap = NULL;
	m_pZBuffer = NULL;
}

CCubeMap::~CCubeMap()
{
	Free();
}

void CCubeMap::Free()
{
	SAFE_RELEASE(m_pCubeMap);
	SAFE_RELEASE(m_pZBuffer);
}

// 
// pD3DDev : [IN] IDirect3DDevice9 IuWFNg
HRESULT CCubeMap::Init( LPDIRECT3DDEVICE9 pD3DDev, UINT iCubemapSize )
{
	if(!pD3DDev)
		return E_FAIL;
	
	Free();
	
	HRESULT hr;
	
	LPDIRECT3DSURFACE9 pBackBuffer = NULL;
	hr = pD3DDev->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &pBackBuffer );
	hr = pBackBuffer->GetDesc( &m_d3dsdBackBuffer );
	hr = pBackBuffer->Release();
	hr = pD3DDev->GetDeviceCaps(&m_d3dCaps);
	
	// L[u}bv쐬
	if( m_d3dCaps.TextureCaps & D3DPTEXTURECAPS_CUBEMAP ){
		if(FAILED(D3DXCreateCubeTexture(pD3DDev, iCubemapSize, 1, D3DUSAGE_RENDERTARGET,
											m_d3dsdBackBuffer.Format, D3DPOOL_DEFAULT, &m_pCubeMap)))
		{
			m_pCubeMap = NULL;
		}
		if(FAILED(pD3DDev->CreateDepthStencilSurface(iCubemapSize, iCubemapSize + 25,
											D3DFMT_D16, D3DMULTISAMPLE_NONE, 0, TRUE, &m_pZBuffer, NULL)))
		{
			m_pZBuffer = NULL;
		}
	}
	else{
		hr = E_FAIL;
	}
	
	return hr;
}

// 6̃L[u}bvɔwi_O
// pD3DDev  : [IN] IDirect3DDevice9 IuWFNg
// v3EyePt  : [IN] _W
// matWorld : [IN] matRY * matT
// matT     : [IN] v3EyePt ̕sړs
// matRY    : [IN] _̉]s
HRESULT CCubeMap::RenderSceneIntoCubeMap( LPDIRECT3DDEVICE9 pD3DDev, D3DXVECTOR3 v3EyePt,
							D3DXMATRIX matWorld, D3DXMATRIX matT, D3DXMATRIX matRY )
{
	if( !m_pCubeMap || !m_pZBuffer )
		return E_FAIL;
	
	HRESULT hr;
	DWORD i;

	// Save transformation matrices of the device
	D3DXMATRIX	 matProjSave, matViewSave;
	hr = pD3DDev->GetTransform( D3DTS_VIEW, 	 &matViewSave );
	hr = pD3DDev->GetTransform( D3DTS_PROJECTION, &matProjSave );
	
	// Set the projection matrix for a field of view of 90 degrees
	D3DXMATRIX matProj;
	D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/2, 1.0f, 0.5f, 100.0f );
	hr = pD3DDev->SetTransform( D3DTS_PROJECTION, &matProj );

	// Store the current backbuffer
	LPDIRECT3DSURFACE9 pBackBuffer,pZBuffer;
	hr = pD3DDev->GetRenderTarget( 0, &pBackBuffer );
	hr = pD3DDev->GetDepthStencilSurface( &pZBuffer );

	// Render to the six faces of the cube map
	pD3DDev->SetDepthStencilSurface( m_pZBuffer );
	for( i=0; i<6; i++ )
	{
		// Set the view transform for this cubemap surface
		D3DXMATRIX matView;
		matView = GetCubeMapViewMatrix( (D3DCUBEMAP_FACES)i, v3EyePt, matWorld, matT, matRY );
		hr = pD3DDev->SetTransform( D3DTS_VIEW, &matView );

		// Set the rendertarget to the i'th cubemap surface
		LPDIRECT3DSURFACE9 pCubeMapFace;
		hr = m_pCubeMap->GetCubeMapSurface( (D3DCUBEMAP_FACES)i, 0, &pCubeMapFace );
		hr = pD3DDev->SetRenderTarget( 0, pCubeMapFace );
		hr = pCubeMapFace->Release();

		// Render the scene (except for the teapot)
		hr = pD3DDev->Clear( 0, NULL, D3DCLEAR_ZBUFFER, 0, 1.0f, 0 );
		hr = pD3DDev->BeginScene();
		hr = RenderScene( pD3DDev, matWorld, matT, matRY );
		hr = pD3DDev->EndScene();
	}

	// Change the rendertarget back to the main backbuffer
	hr = pD3DDev->SetRenderTarget( 0, pBackBuffer );
	hr = pD3DDev->SetDepthStencilSurface( pZBuffer );
	hr = pBackBuffer->Release();
	hr = pZBuffer->Release();

	// Restore the original transformation matrices
	hr = pD3DDev->SetTransform( D3DTS_VIEW,    &matViewSave );
	hr = pD3DDev->SetTransform( D3DTS_PROJECTION, &matProjSave );

	return hr;
}

// w̃r[s擾
// dwFace   : [IN] L[u}bv̖ʔԍ
// v3EyePt  : [IN] _W
// matWorld : [IN] matRY * matT
// matT     : [IN] v3EyePt ̕sړs
// matRY    : [IN] _Y]s
D3DXMATRIX CCubeMap::GetCubeMapViewMatrix( DWORD dwFace, D3DXVECTOR3 v3EyePt,
							D3DXMATRIX matWorld, D3DXMATRIX matT, D3DXMATRIX matRY )
{
	D3DXVECTOR3 v3LookDir;
	D3DXVECTOR3 v3UpDir;

	switch( dwFace )
	{
		case D3DCUBEMAP_FACE_POSITIVE_X:
			v3LookDir = D3DXVECTOR3( 1.0f, 0.0f, 0.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
			break;
		case D3DCUBEMAP_FACE_NEGATIVE_X:
			v3LookDir = D3DXVECTOR3(-1.0f, 0.0f, 0.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
			break;
		case D3DCUBEMAP_FACE_POSITIVE_Y:
			v3LookDir = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 0.0f,-1.0f );
			break;
		case D3DCUBEMAP_FACE_NEGATIVE_Y:
			v3LookDir = D3DXVECTOR3( 0.0f,-1.0f, 0.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 0.0f, 1.0f );
			break;
		case D3DCUBEMAP_FACE_POSITIVE_Z:
			v3LookDir = D3DXVECTOR3( 0.0f, 0.0f, 1.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
			break;
		case D3DCUBEMAP_FACE_NEGATIVE_Z:
			v3LookDir = D3DXVECTOR3( 0.0f, 0.0f,-1.0f );
			v3UpDir   = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
			break;
	}

	switch( dwFace )
	{
		case D3DCUBEMAP_FACE_POSITIVE_X:
		case D3DCUBEMAP_FACE_NEGATIVE_X:
		case D3DCUBEMAP_FACE_POSITIVE_Z:
		case D3DCUBEMAP_FACE_NEGATIVE_Z:
			D3DXVec3TransformCoord(&v3LookDir, &v3LookDir, &matWorld);
			break;
		case D3DCUBEMAP_FACE_POSITIVE_Y:
		case D3DCUBEMAP_FACE_NEGATIVE_Y:
			D3DXVec3TransformCoord(&v3UpDir, &v3UpDir, &matRY);
			D3DXVec3TransformCoord(&v3LookDir, &v3LookDir, &matT);
			break;
	}

	// Set the view transform for this cubemap surface
	D3DXMATRIX matView;
	D3DXMatrixLookAtLH( &matView, &v3EyePt, &v3LookDir, &v3UpDir );
	return matView;
}

#endif