// Xt@C\
// l^FDirectX9 Tv d3dfile.cpp
#ifndef _FILE_X_H_
#define _FILE_X_H_

#include <d3dx9.h>
#include "macro.h"

class CFileX
{
public:
	CFileX();
	~CFileX();

	// Creation/destruction
	HRESULT Load( LPDIRECT3DDEVICE9 pD3DDev, char* strFilename );
	void	Free();

	// Rendering
	HRESULT Render( LPDIRECT3DDEVICE9 pD3DDev, bool bMaterial = true,
					bool bTexture = true );

	// Mesh access
	LPD3DXMESH GetMesh() { return m_pMesh; }

	// _tH[}bg擾
	DWORD GetFVF() { return m_pMesh->GetFVF(); }

	// _tH[}bgύX
	HRESULT SetFVF( LPDIRECT3DDEVICE9 pD3DDev, DWORD dwFVF );

	// tO擾
	bool	GetFlagAlpha() { return m_bAlpha; }

private:
	bool				m_bAlpha;
	LPD3DXMESH			m_pMesh;
	DWORD				m_dwNumMaterials;
	D3DMATERIAL9*		m_pMaterials;
	LPDIRECT3DTEXTURE9* m_pTextures;

};

CFileX::CFileX()
{
	m_bAlpha		 = false;
	m_pMesh 		 = NULL;
	m_dwNumMaterials = 0L;
	m_pMaterials	 = NULL;
	m_pTextures 	 = NULL;
}

CFileX::~CFileX()
{
	Free();
}

// Xt@Cǂݍ
// strFilename : [IN] t@C
HRESULT CFileX::Load( LPDIRECT3DDEVICE9 pD3DDev, char* strFilename )
{
	if(!pD3DDev)
		return E_FAIL;

	Free();

	LPD3DXBUFFER pAdjBuf = NULL;
	LPD3DXBUFFER pD3DXBuf = NULL;
	char		 *p,path[MAX_PATH],path2[MAX_PATH];
	DWORD		 i;
	HRESULT 	 hr;

	// Load the mesh
	strcpy(path, strFilename);
	p = strrchr(path, '\\');
	if(p)
		strcpy(p, "\\");
	else
		strcpy(path, ".\\");
	if( FAILED( hr = D3DXLoadMeshFromX( strFilename, D3DXMESH_MANAGED, pD3DDev,
										&pAdjBuf, &pD3DXBuf, NULL,
										&m_dwNumMaterials, &m_pMesh ) ) )
	{
		return hr;
	}

	// Optimize the mesh for performance
	if( FAILED( hr = m_pMesh->OptimizeInplace(
						D3DXMESHOPT_COMPACT | D3DXMESHOPT_ATTRSORT |
						D3DXMESHOPT_VERTEXCACHE | D3DXMESHOPT_DEVICEINDEPENDENT,
						(DWORD*)pAdjBuf->GetBufferPointer(), NULL, NULL, NULL ) ) )
	{
		goto End;
	}

	// AFVF ɖ@܂܂ĂA@ݒ
	D3DXComputeNormals( m_pMesh, NULL );

	// Get material info for the mesh
	// Get the array of materials out of the buffer
	if( pD3DXBuf && m_dwNumMaterials > 0 )
	{
		// Allocate memory for the materials and textures
		D3DXMATERIAL* D3DXMtrls = (D3DXMATERIAL*)pD3DXBuf->GetBufferPointer();
		m_pMaterials = new D3DMATERIAL9[m_dwNumMaterials];
		if( m_pMaterials == NULL ){
			hr = E_OUTOFMEMORY;
			goto End;
		}
		m_pTextures  = new LPDIRECT3DTEXTURE9[m_dwNumMaterials];
		if( m_pTextures == NULL ){
			hr = E_OUTOFMEMORY;
			goto End;
		}

		// Copy each material and create its texture
		for( i=0; i<m_dwNumMaterials; i++ ){
			// Copy the material
			m_pMaterials[i] = D3DXMtrls[i].MatD3D;
			m_pMaterials[i].Ambient = m_pMaterials[i].Diffuse;
			m_pTextures[i]	= NULL;
			if(m_pMaterials[i].Diffuse.a < 1.0f){
				m_bAlpha = true;
			}

			// Create a texture
			if( D3DXMtrls[i].pTextureFilename ){
				strcpy(path2, path);
				strcat(path2, D3DXMtrls[i].pTextureFilename);
				if( FAILED( D3DXCreateTextureFromFile( pD3DDev, path2, 
													   &m_pTextures[i] ) ) )
					m_pTextures[i] = NULL;
			}
		}
	}
	hr = S_OK;

End:
	SAFE_RELEASE( pAdjBuf );
	SAFE_RELEASE( pD3DXBuf );

	return hr;
}

// 
void CFileX::Free()
{
	UINT i;
	for( i=0; i<m_dwNumMaterials; i++ )
		SAFE_RELEASE( m_pTextures[i] );
	SAFE_DELETE_ARRAY( m_pTextures );
	SAFE_DELETE_ARRAY( m_pMaterials );
	SAFE_RELEASE( m_pMesh );
	m_dwNumMaterials = 0L;
	m_bAlpha		 = false;
}

// bV`
// pD3DDev	 : [IN] IDirect3DDevice9 IuWFNg
// bMaterial : [IN] }eAKptO
// bTexture  : [IN] eLX`KptO
HRESULT CFileX::Render( LPDIRECT3DDEVICE9 pD3DDev, bool bMaterial, bool bTexture )
{
	if( !pD3DDev || !m_pMesh )
		return E_FAIL;

	// Frist, draw the subsets without alpha
	DWORD i;
	for( i=0; i<m_dwNumMaterials; i++ ){
		if( m_pMaterials[i].Diffuse.a < 1.0f )
			continue;
		if( bMaterial )
			pD3DDev->SetMaterial( &m_pMaterials[i] );
		if( bTexture )
			pD3DDev->SetTexture( 0, m_pTextures[i] );
		m_pMesh->DrawSubset( i );
	}

	// Then, draw the subsets with alpha
	for( i=0; i<m_dwNumMaterials; i++ ){
		if( m_pMaterials[i].Diffuse.a == 1.0f )
			continue;
		if( bMaterial )
			pD3DDev->SetMaterial( &m_pMaterials[i] );
		if( bTexture )
			pD3DDev->SetTexture( 0, m_pTextures[i] );
		m_pMesh->DrawSubset( i );
	}

	return S_OK;
}

// _tH[}bgύX
HRESULT CFileX::SetFVF( LPDIRECT3DDEVICE9 pD3DDev, DWORD dwFVF )
{
	LPD3DXMESH pTempMesh  = NULL;

	if( m_pMesh ){
		if( FAILED( m_pMesh->CloneMeshFVF( 0L, dwFVF, pD3DDev, &pTempMesh ) ) ){
			return E_FAIL;
		}
	}

	SAFE_RELEASE( m_pMesh );

	if( pTempMesh )  m_pMesh  = pTempMesh;

	// Compute normals in case the meshes have them
	if( m_pMesh )
		D3DXComputeNormals( m_pMesh, NULL );

	return S_OK;
}

#endif