// Xt@C Aj\
// l^FDirectX9 Tv
#ifndef _FILE_XANIM_H_
#define _FILE_XANIM_H_

#include <d3dx9.h>
#include "macro.h"
#include "meshcontainer.h"

#pragma comment(lib,"winmm.lib")

class CFileXAnim
{
public:
	CFileXAnim();
	~CFileXAnim();

	// Creation/destruction
	HRESULT Load( LPDIRECT3DDEVICE9 pD3DDev, char* strFilename );
	void	Free();

	// Rendering
	HRESULT Render(LPDIRECT3DDEVICE9 pD3DDev, bool bMaterial = true,
					bool bTexture = true);

	// Data access
	LPD3DXANIMATIONCONTROLLER GetAnimController() { return m_pAnimController; }
	int 					  GetNumAnimSet()	  { return m_numAnimSet; }
	int 					  GetIndexAnimeSet()  { return m_idxAnimeSet; }
	bool					  GetFlagAlpha()	  { return m_bAlpha; }
	DWORD					  GetAnimIndexFromName(char *strAnimname);

private:
	D3DCAPS9					m_d3dCaps;
	CAllocateHierarchy* 		m_pAlloc;
	LPD3DXFRAME 				m_pFrameRoot;
	LPD3DXANIMATIONCONTROLLER	m_pAnimController;
	int 						m_numAnimSet;	// Aj[VZbg
	int 						m_idxAnimeSet;	// ݑIĂAj[VZbgԍ
	bool						m_bAlpha;		// tO

private:
	HRESULT SetupBoneMatrixPointers(LPD3DXFRAME pFrame);
	HRESULT SetupBoneMatrixPointersOnMesh(LPD3DXMESHCONTAINER pMeshContainerBase);
	HRESULT DrawFrame(LPDIRECT3DDEVICE9 pD3DDev, LPD3DXFRAME pFrame,
						bool bMaterial, bool bTexture);
	void	DrawMeshContainer(LPDIRECT3DDEVICE9 pD3DDev,
								LPD3DXMESHCONTAINER pMeshContainerBase,
								LPD3DXFRAME pFrameBase,
								bool bMaterial, bool bTexture);
	HRESULT FrameMove(LPDIRECT3DDEVICE9 pD3DDev, D3DXMATRIX matWorld);
	void	UpdateFrameMatrices(LPD3DXFRAME pFrameBase, LPD3DXMATRIX pParentMatrix);
};

CFileXAnim::CFileXAnim()
{
	ZeroMemory(&m_d3dCaps, sizeof(m_d3dCaps));
	m_pAlloc		  = NULL;
	m_pFrameRoot	  = NULL;
	m_pAnimController = NULL;
	m_numAnimSet	  = 0;
	m_idxAnimeSet	  = 0;
	m_bAlpha		  = false;
}

CFileXAnim::~CFileXAnim()
{
	Free();
}

// Xt@Cǂݍ
// strFilename : [IN] t@C
HRESULT CFileXAnim::Load( LPDIRECT3DDEVICE9 pD3DDev, char* strFilename )
{
	if(!pD3DDev)
		return E_FAIL;
	
	Free();
	
	UINT i;
	
	pD3DDev->GetDeviceCaps(&m_d3dCaps);
	
	m_pAlloc = new CAllocateHierarchy;
	
	char *p,path[MAX_PATH];
	strcpy(path, strFilename);
	p = strrchr(path, '\\');
	if(p)
		strcpy(p, "\\");
	m_pAlloc->Set_Path(path);
	
	// X t@Cǂݍ
	if(FAILED(D3DXLoadMeshHierarchyFromX(strFilename, D3DXMESH_MANAGED, pD3DDev, m_pAlloc, NULL, &m_pFrameRoot, &m_pAnimController)))
		return E_FAIL;
	
	// {[sݒ
	if(FAILED(SetupBoneMatrixPointers(m_pFrameRoot)))
		return E_FAIL;
	
	// gbNɃAj[VZbg
	LPD3DXANIMATIONSET	pAnimSet;
	D3DXTRACK_DESC		Desc;
	ZeroMemory(&Desc, sizeof(Desc));
	Desc.Weight = 1;
	Desc.Speed	= 1;
	Desc.Enable = FALSE;
	m_numAnimSet = m_pAnimController->GetNumAnimationSets();
	for(i=1; i<(UINT)m_numAnimSet; i++){
		m_pAnimController->GetAnimationSet(i, &pAnimSet);
		m_pAnimController->SetTrackAnimationSet(i, pAnimSet);
		m_pAnimController->SetTrackDesc(i, &Desc);
		SAFE_RELEASE(pAnimSet);
	}
	m_pAnimController->SetTrackEnable(m_idxAnimeSet, TRUE);
	
	LPD3DXMESHCONTAINER pMeshContainer;
	pMeshContainer = m_pFrameRoot->pMeshContainer;
	while (pMeshContainer != NULL)
	{
		for(i=0; i<pMeshContainer->NumMaterials; i++){
			if(pMeshContainer->pMaterials[i].MatD3D.Diffuse.a < 1.0f){
				m_bAlpha = true;
				break;
			}
		}
		pMeshContainer = pMeshContainer->pNextMeshContainer;
	}
	
	return S_OK;
}

// 
void CFileXAnim::Free()
{
	SAFE_RELEASE(m_pAnimController);
	D3DXFrameDestroy(m_pFrameRoot, m_pAlloc);
	SAFE_DELETE(m_pAlloc);
	m_numAnimSet	  = 0;
	m_idxAnimeSet	  = 0;
	m_bAlpha		  = false;
}

// bV`
HRESULT CFileXAnim::Render(LPDIRECT3DDEVICE9 pD3DDev, bool bMaterial, bool bTexture)
{
	HRESULT 				hr;
	LPDIRECT3DVERTEXSHADER9 dwVS_pre;
	DWORD					dwFVF_pre,dwVB_pre;
	D3DXMATRIX				mat1_pre,mat2_pre;
	
	hr = pD3DDev->GetVertexShader(&dwVS_pre);
	hr = pD3DDev->GetFVF(&dwFVF_pre);
	hr = pD3DDev->GetRenderState(D3DRS_VERTEXBLEND, &dwVB_pre);
	hr = pD3DDev->GetTransform(D3DTS_WORLDMATRIX(0), &mat1_pre);
	hr = pD3DDev->GetTransform(D3DTS_WORLDMATRIX(1), &mat2_pre);
	hr = FrameMove(pD3DDev, mat1_pre);
	if(SUCCEEDED(hr))
		hr = DrawFrame(pD3DDev, m_pFrameRoot, bMaterial, bTexture);
	
	hr = pD3DDev->SetVertexShader(dwVS_pre);
	hr = pD3DDev->SetFVF(dwFVF_pre);
	hr = pD3DDev->SetRenderState(D3DRS_VERTEXBLEND, dwVB_pre);
	hr = pD3DDev->SetTransform(D3DTS_WORLDMATRIX(0), &mat1_pre);
	hr = pD3DDev->SetTransform(D3DTS_WORLDMATRIX(1), &mat2_pre);
	
	return hr;
}

// AjAjԍ擾
DWORD CFileXAnim::GetAnimIndexFromName(char *strAnimname)
{
    LPD3DXANIMATIONSET	pAS;
    DWORD				i,index = -1;

    for(i=0; i<m_pAnimController->GetNumAnimationSets(); ++i)
    {
        if(FAILED(m_pAnimController->GetAnimationSet(i, &pAS)))
            continue;

        if(pAS->GetName() &&
			!strnicmp(pAS->GetName(), strAnimname, strlen(pAS->GetName())))
        {
            index = i;
            pAS->Release();
            break;
        }
        pAS->Release();
    }

    return index;
}

HRESULT CFileXAnim::SetupBoneMatrixPointers(LPD3DXFRAME pFrame)
{
	HRESULT hr;

	if (pFrame->pMeshContainer != NULL)
	{
		hr = SetupBoneMatrixPointersOnMesh(pFrame->pMeshContainer);
		if (FAILED(hr))
			return hr;
	}

	if (pFrame->pFrameSibling != NULL)
	{
		hr = SetupBoneMatrixPointers(pFrame->pFrameSibling);
		if (FAILED(hr))
			return hr;
	}

	if (pFrame->pFrameFirstChild != NULL)
	{
		hr = SetupBoneMatrixPointers(pFrame->pFrameFirstChild);
		if (FAILED(hr))
			return hr;
	}

	return S_OK;
}// SetupBoneMatrixPointers

HRESULT CFileXAnim::SetupBoneMatrixPointersOnMesh(LPD3DXMESHCONTAINER pMeshContainerBase)
{
	UINT iBone, cBones;
	D3DXFRAME_DERIVED *pFrame;

	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;

	// if there is a skinmesh, then setup the bone matrices
	if (pMeshContainer->pSkinInfo != NULL)
	{
		cBones = pMeshContainer->pSkinInfo->GetNumBones();

		pMeshContainer->ppBoneMatrixPtrs = new D3DXMATRIX*[cBones];
		if (pMeshContainer->ppBoneMatrixPtrs == NULL)
			return E_OUTOFMEMORY;

		for (iBone = 0; iBone < cBones; iBone++)
		{
			pFrame = (D3DXFRAME_DERIVED*)D3DXFrameFind(m_pFrameRoot, pMeshContainer->pSkinInfo->GetBoneName(iBone));
			if (pFrame == NULL)
				return E_FAIL;

			pMeshContainer->ppBoneMatrixPtrs[iBone] = &pFrame->CombinedTransformationMatrix;
		}
	}

	return S_OK;
}// SetupBoneMatrixPointersOnMesh

HRESULT CFileXAnim::DrawFrame(LPDIRECT3DDEVICE9 pD3DDev, LPD3DXFRAME pFrame,
								bool bMaterial, bool bTexture)
{
	if(!pD3DDev || !pFrame)
		return E_FAIL;

	LPD3DXMESHCONTAINER pMeshContainer;

	pMeshContainer = pFrame->pMeshContainer;
	while (pMeshContainer != NULL)
	{
		DrawMeshContainer(pD3DDev, pMeshContainer, pFrame, bMaterial, bTexture);

		pMeshContainer = pMeshContainer->pNextMeshContainer;
	}

	if (pFrame->pFrameSibling != NULL)
	{
		DrawFrame(pD3DDev, pFrame->pFrameSibling, bMaterial, bTexture);
	}

	if (pFrame->pFrameFirstChild != NULL)
	{
		DrawFrame(pD3DDev, pFrame->pFrameFirstChild, bMaterial, bTexture);
	}

	return S_OK;
}// DrawFrame

void CFileXAnim::DrawMeshContainer(LPDIRECT3DDEVICE9 pD3DDev,
									LPD3DXMESHCONTAINER pMeshContainerBase,
									LPD3DXFRAME pFrameBase,
									bool bMaterial, bool bTexture)
{
	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;
	D3DXFRAME_DERIVED *pFrame = (D3DXFRAME_DERIVED*)pFrameBase;
	UINT iMaterial;
	UINT NumBlend;
	UINT iAttrib;
	DWORD AttribIdPrev;
	LPD3DXBONECOMBINATION pBoneComb;

	UINT iMatrixIndex;
	D3DXMATRIXA16 matTemp;

	// first check for skinning
	if (pMeshContainer->pSkinInfo != NULL)
	{
		AttribIdPrev = UNUSED32; 
		pBoneComb = reinterpret_cast<LPD3DXBONECOMBINATION>(pMeshContainer->pBoneCombinationBuf->GetBufferPointer());

		// Draw using default vtx processing of the device (typically HW)
		for (iAttrib = 0; iAttrib < pMeshContainer->NumAttributeGroups; iAttrib++)
		{
			NumBlend = 0;
			for (DWORD i = 0; i < pMeshContainer->NumInfl; ++i)
			{
				if (pBoneComb[iAttrib].BoneId[i] != UINT_MAX)
				{
					NumBlend = i;
				}
			}

			if (m_d3dCaps.MaxVertexBlendMatrices >= NumBlend + 1)
			{
				// first calculate the world matrices for the current set of blend weights and get the accurate count of the number of blends
				for (DWORD i = 0; i < pMeshContainer->NumInfl; ++i)
				{
					iMatrixIndex = pBoneComb[iAttrib].BoneId[i];
					if (iMatrixIndex != UINT_MAX)
					{
						D3DXMatrixMultiply( &matTemp, &pMeshContainer->pBoneOffsetMatrices[iMatrixIndex], pMeshContainer->ppBoneMatrixPtrs[iMatrixIndex] );
						pD3DDev->SetTransform( D3DTS_WORLDMATRIX( i ), &matTemp );
					}
				}

				pD3DDev->SetRenderState(D3DRS_VERTEXBLEND, NumBlend);

				// lookup the material used for this subset of faces
				if ((AttribIdPrev != pBoneComb[iAttrib].AttribId) || (AttribIdPrev == UNUSED32))
				{
					if(bMaterial)
						pD3DDev->SetMaterial( &pMeshContainer->pMaterials[pBoneComb[iAttrib].AttribId].MatD3D );
					if(bTexture)
						pD3DDev->SetTexture( 0, pMeshContainer->ppTextures[pBoneComb[iAttrib].AttribId] );
					AttribIdPrev = pBoneComb[iAttrib].AttribId;
				}

				// draw the subset now that the correct material and matrices are loaded
				pMeshContainer->MeshData.pMesh->DrawSubset(iAttrib);
			}
		}
		pD3DDev->SetRenderState(D3DRS_VERTEXBLEND, 0);
	}
	else  // standard mesh, just draw it after setting material properties
	{
		pD3DDev->SetTransform(D3DTS_WORLD, &pFrame->CombinedTransformationMatrix);
		for (iMaterial = 0; iMaterial < pMeshContainer->NumMaterials; iMaterial++)
		{
			if(bMaterial)
				pD3DDev->SetMaterial( &pMeshContainer->pMaterials[iMaterial].MatD3D );
			if(bTexture)
				pD3DDev->SetTexture( 0, pMeshContainer->ppTextures[iMaterial] );
			pMeshContainer->MeshData.pMesh->DrawSubset(iMaterial);
		}
	}
}// DrawMeshContainer

HRESULT CFileXAnim::FrameMove(LPDIRECT3DDEVICE9 pD3DDev, D3DXMATRIX matWorld)
{
	if(!pD3DDev || !m_pAnimController || !m_pFrameRoot)
		return E_FAIL;

	if (m_pAnimController != NULL){
		float			pass_seconds = 0;
		DWORD			now_time;
		static DWORD	pre_time = 0;

		now_time = timeGetTime();
		if(pre_time)
			pass_seconds = (float)(now_time - pre_time) / 1000.0f;
		pre_time = now_time;
		m_pAnimController->SetTime(m_pAnimController->GetTime() + pass_seconds);
	}

	pD3DDev->SetTransform(D3DTS_WORLD, &matWorld);
	
	UpdateFrameMatrices(m_pFrameRoot, &matWorld);

	return S_OK;
}// FrameMove

void CFileXAnim::UpdateFrameMatrices(LPD3DXFRAME pFrameBase, LPD3DXMATRIX pParentMatrix)
{
	D3DXFRAME_DERIVED *pFrame = (D3DXFRAME_DERIVED*)pFrameBase;

	if (pParentMatrix != NULL)
		D3DXMatrixMultiply(&pFrame->CombinedTransformationMatrix, &pFrame->TransformationMatrix, pParentMatrix);
	else
		pFrame->CombinedTransformationMatrix = pFrame->TransformationMatrix;

	if (pFrame->pFrameSibling != NULL)
	{
		UpdateFrameMatrices(pFrame->pFrameSibling, pParentMatrix);
	}

	if (pFrame->pFrameFirstChild != NULL)
	{
		UpdateFrameMatrices(pFrame->pFrameFirstChild, &pFrame->CombinedTransformationMatrix);
	}
}// UpdateFrameMatrices

#endif